import pytest
import stribor as st
import torch
import torch.nn as nn

class ModelExact(nn.Module):
    """
    The same forward function as defined in `stribor.net.FuncAndDiagJac`
    but without detaching and without custom gradients.
    """
    def __init__(self, exclusive_net, dimwise_net):
        super().__init__()
        self.exclusive_net = exclusive_net
        self.dimwise_net = dimwise_net

    def forward(self, t, x, latent=None):
        shape = x.shape
        h = self.exclusive_net(t, x)
        x = x.view(-1, 1)
        h = h.reshape(x.shape[0], -1)
        if latent is not None:
            latent = latent.unsqueeze(-2).repeat_interleave(shape[-1], -2)
            h = torch.cat([h, latent.reshape(x.shape[0], -1)], -1)
        output = self.dimwise_net(t, x, h)
        return output.view(*shape)

def check_if_model_is_implemented_correctly(input_shape, model, latent_dim):
    # Define the exact model to which we are comparing with
    model_exact = ModelExact(model.exclusive_net, model.dimwise_net)

    # Input data
    t = torch.Tensor([1])
    x = torch.randn(input_shape)
    if latent_dim == 0:
        latent = None
    else:
        latent = torch.randn(*x.shape[:-1], latent_dim)

    # Computed output and Jacobian diagonal
    y, jac = model(t, x, latent=latent)

    # Exact output and Jacobian diagonal
    t_ = t.clone().requires_grad_(True)
    x_ = x.clone().requires_grad_(True)
    if latent is None:
        f = lambda t, x: model_exact(t, x, None)
        y_exact = f(t_, x_)
        jac_exact = st.util.divergence_from_jacobian(f, (t_, x_))[1]
    else:
        latent_ = latent.clone().requires_grad_(True)
        f = lambda t, x, latent: model_exact(t, x, latent)
        y_exact = f(t_, x_, latent_)
        jac_exact = st.util.divergence_from_jacobian(f, (t_, x_, latent_))[1]

    # Check if forward pass and jacobian are correct
    assert torch.allclose(y, y_exact), 'Output is incorrect'
    assert torch.allclose(jac, jac_exact), 'Jacobian is incorrect'

    # Simulate loss and test if backward step works
    loss1 = y.mean()
    loss1.backward()

    loss2 = y_exact.mean()
    loss2.backward()

    # Check if losses for both models are exactly the same
    assert all([torch.allclose(g1.grad, g2.grad) for g1, g2 in zip(model.parameters(), model_exact.parameters())]), \
        'Gradients for two networks are not the same'

@pytest.mark.parametrize('input_shape', [(10, 2), (1, 1, 3), (3, 7, 8), (2, 3, 5, 2)])
@pytest.mark.parametrize('hidden_dims', [[], [64, 32]])
@pytest.mark.parametrize('d_h', [1, 2, 5])
@pytest.mark.parametrize('latent_dim', [0, 1, 32])
def test_diffeq_exact_trace_mlp(input_shape, hidden_dims, d_h, latent_dim):
    torch.manual_seed(123)
    model = st.net.DiffeqExactTraceMLP(input_shape[-1], hidden_dims, input_shape[-1], d_h, latent_dim=latent_dim)
    check_if_model_is_implemented_correctly(input_shape, model, latent_dim)


@pytest.mark.parametrize('input_shape', [(3, 7, 2), (2, 3, 5, 2)])
@pytest.mark.parametrize('hidden_dims', [[], [64, 32]])
@pytest.mark.parametrize('d_h', [1, 2, 5])
@pytest.mark.parametrize('pooling', ['max', 'mean'])
@pytest.mark.parametrize('latent_dim', [0, 1, 32])
def test_diffeq_exact_trace_deepset(input_shape, hidden_dims, d_h, latent_dim, pooling):
    torch.manual_seed(123)
    model = st.net.DiffeqExactTraceDeepSet(input_shape[-1], hidden_dims, input_shape[-1], d_h,
                                           latent_dim=latent_dim, pooling=pooling)
    check_if_model_is_implemented_correctly(input_shape, model, latent_dim)


@pytest.mark.parametrize('input_shape', [(1, 2, 8), (3, 7, 2), (2, 3, 5, 2)])
@pytest.mark.parametrize('hidden_dims', [[64, 32]])
@pytest.mark.parametrize('d_h', [1, 3, 7])
@pytest.mark.parametrize('n_heads', [1, 2, 8])
@pytest.mark.parametrize('latent_dim', [0, 1, 32])
def test_diffeq_exact_trace_attention(input_shape, hidden_dims, d_h, latent_dim, n_heads):
    torch.manual_seed(123)
    model = st.net.DiffeqExactTraceAttention(input_shape[-1], hidden_dims, input_shape[-1], d_h,
                                             latent_dim=latent_dim, n_heads=n_heads)
    check_if_model_is_implemented_correctly(input_shape, model, latent_dim)


@pytest.mark.parametrize('input_shape', [(1, 2, 8), (3, 7, 2), (2, 3, 5, 2), (3, 16)])
@pytest.mark.parametrize('hidden_dims', [[32, 64]])
@pytest.mark.parametrize('d_h', [1, 3, 7])
@pytest.mark.parametrize('latent_dim', [32])
def test_backpropagation_through_encoder(input_shape, hidden_dims, d_h, latent_dim):
    class Model(nn.Module):
        def __init__(self, diffeq):
            super().__init__()
            self.diffeq = diffeq
            self.encoder = st.net.MLP(input_shape[-1], [17], latent_dim)

        def forward(self, t, x):
            latent = self.encoder(torch.ones(input_shape)).requires_grad_(True)
            y = self.diffeq(t, x, latent=latent)
            return y

    torch.manual_seed(123)
    t = torch.Tensor([1])
    x = torch.randn(*input_shape)
    diffeq = st.net.DiffeqExactTraceMLP(input_shape[-1], hidden_dims, input_shape[-1], d_h,
                                      latent_dim=latent_dim, return_log_det_jac=False)
    model = Model(diffeq)
    y = model(t, x)
    loss = y.mean()
    loss.backward()

    torch.manual_seed(123)
    t_ = torch.ones(1).requires_grad_(True)
    x_ = torch.randn(*input_shape).requires_grad_(True)
    diffeq_exact = st.net.DiffeqExactTraceMLP(input_shape[-1], hidden_dims, input_shape[-1], d_h,
                                            latent_dim=latent_dim, return_log_det_jac=False)
    diffeq_exact = ModelExact(diffeq.exclusive_net, diffeq.dimwise_net)
    model_exact = Model(diffeq_exact)
    y_exact = model_exact(t_, x_)
    loss_exact = y_exact.mean()
    loss_exact.backward()

    assert torch.isclose(y, y_exact).all(), 'Outputs are not the same'

    # Check if gradients for both models are exactly the same
    for (n1, g1), (n2, g2) in zip(model.named_parameters(), model_exact.named_parameters()):
        assert n1 == n2, 'Names are not the same'
        assert (g1.data == g2.data).all(), 'Parameter values are not the same'
        assert torch.isclose(g1.grad, g2.grad).all(), f'Gradients are not the same'



import pytest
import torch
import stribor as st

@pytest.mark.parametrize('input_dim', [(10, 2), (2, 10), (1, 10, 2), (5, 10, 2)])
@pytest.mark.parametrize('natural_ordering', [True, False])
@pytest.mark.parametrize('reverse_ordering', [True, False])
@pytest.mark.parametrize('return_per_dim', [True, False])
def test_made_zero_trace(input_dim, natural_ordering, reverse_ordering, return_per_dim):
    torch.manual_seed(123)

    dim = input_dim[-1]
    model = st.net.MADE(
        in_dim=dim,
        hidden_dims=[64, 64],
        out_dim=dim * 8,
        natural_ordering=natural_ordering,
        reverse_ordering=reverse_ordering,
        return_per_dim=return_per_dim
    )

    x = torch.randn(*input_dim).requires_grad_(True)

    f = lambda x: model(x)
    trace_exact = st.util.divergence_from_jacobian(f, x)

    assert trace_exact.sum() == 0



import torch
import pytest
import numpy as np
import stribor as st
from stribor.test.base import *


@pytest.mark.parametrize('input_shape', [(2, 10), (7, 4, 5)])
@pytest.mark.parametrize('latent_dim', [0, 1, 13])
def test_cnf_computation(input_shape, latent_dim):
    np.random.seed(123)
    torch.manual_seed(123)

    x = torch.rand(*input_shape)
    latent = torch.randn(*x.shape[:-1], latent_dim) if latent_dim > 0 else None

    dim = x.shape[-1]
    f = st.ContinuousTransform(
        dim,
        net=st.net.DiffeqMLP(dim + latent_dim + 1, [32], dim),
        atol=1e-8,
        rtol=1e-8,
        divergence='compute',
        solver='dopri5',
        has_latent=latent is not None,
    )

    check_inverse_transform(f, x, latent=latent)
    check_log_jacobian_determinant(f, x, latent=latent)
    check_gradients_not_nan(f, x, latent=latent)


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('latent_dim', [0, 1, 32])
@pytest.mark.parametrize('diffeq', ['DiffeqMLP'])
@pytest.mark.parametrize('hidden_dims', [[], [32], [32, 64]])
@pytest.mark.parametrize('solver,solver_options', [('dopri5', None), ('rk4', {'solver_step': 0.1})])
@pytest.mark.parametrize('rademacher', [True, False])
@pytest.mark.parametrize('use_adjoint', [True, False])
def test_cnf_definition(input_shape, latent_dim, diffeq, hidden_dims, solver, solver_options, rademacher, use_adjoint):
    np.random.seed(123)
    torch.manual_seed(123)

    x = torch.rand(*input_shape)
    latent = torch.randn(*x.shape[:-1], latent_dim) if latent_dim > 0 else None

    dim = x.shape[-1]
    f = st.ContinuousTransform(
        dim,
        net=getattr(st.net, diffeq)(dim + latent_dim + 1, hidden_dims, dim),
        solver=solver,
        solver_options=solver_options,
        rademacher=rademacher,
        use_adjoint=use_adjoint,
        has_latent=latent is not None,
    )

    check_gradients_not_nan(f, x, latent=latent)


@pytest.mark.parametrize('input_shape', [(1, 1, 1), (3, 4, 5), (2, 3, 4, 5)])
@pytest.mark.parametrize('odenet', ['DiffeqDeepset', 'DiffeqSelfAttention'])
def test_diffeq_equivariant(input_shape, odenet):
    torch.manual_seed(123)

    x = torch.rand(*input_shape)

    dim = input_shape[-1]
    f = st.ContinuousTransform(
        dim,
        net=getattr(st.net, odenet)(dim + 1, [13], dim),
        atol=1e-8,
        rtol=1e-8,
        divergence='compute',
        solver='dopri5',
    )

    check_inverse_transform(f, x)
    check_gradients_not_nan(f, x)



from torchtyping import TensorType
import warnings

import torch
import torch.nn as nn
import torch.nn.functional as F

from stribor import ElementwiseTransform
from stribor.flow import Transform


def diff(x: TensorType[..., 'dim'], dim: int = -1) -> TensorType[..., 'dim']:
    """
    Inverse of torch.cumsum(x, dim=dim).
    Compute differences between subsequent elements of the tensor.
    Only works on dims -1 and -2.

    Args:
        x (tensor): Input of arbitrary shape

    Returns:
        diff (tensor): Result with the same shape as x
    """
    if dim == 1:
        if x.dim() == 2:
            dim = -1
        elif x.dim() == 3:
            dim = -2
        else:
            raise ValueError('If dim=1, tensor must have 2 or 3 dimensions')

    if dim == 2:
        if x.dim() == 3:
            dim = -1
        elif x.dim() == 4:
            dim = -2
        else:
            raise ValueError('If dim=2, tensor should have 3 or 4 dimensions')

    if dim == -1:
        return x - F.pad(x, (1, 0))[..., :-1]
    elif dim == -2:
        return x - F.pad(x, (0, 0, 1, 0))[..., :-1, :]
    else:
        raise ValueError("dim must be equal to -1 or -2")


class Cumsum(ElementwiseTransform):
    """
    Cumulative sum along the specified dimension of the tensor.

    Example:
    >>> f = stribor.Cumsum(-1)
    >>> f(torch.ones(1, 4))
    (tensor([[1., 2., 3., 4.]]), tensor([[0., 0., 0., 0.]]))

    Args:
        dim (int): Tensor dimension over which to perform the summation. Options: -1 or -2.
    """
    def __init__(self, dim: int):
        super().__init__()
        assert dim == -1, '`dim` must be equal to -1'
        self.dim = dim

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        y = x.cumsum(self.dim)
        return y

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        x = diff(y, self.dim)
        return x

    def log_det_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 1]:
        return torch.zeros_like(x[...,:1])

    def log_diag_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        return torch.zeros_like(x)

class Diff(Cumsum):
    """
    Inverse of Cumsum transformation.
    """
    def forward(self, x, **kwargs):
        return super().inverse(x, **kwargs)

    def inverse(self, y, **kwargs):
        return super().forward(y, **kwargs)


class CumsumColumn(Transform):
    """
    Cumulative sum along the specific *single* column in (..., M, N) matrix.

    Example:
    >>> f = stribor.CumsumColumn(1)
    >>> f(torch.ones(3, 3))[0]
    tensor([[1., 1., 1.],
            [1., 2., 1.],
            [1., 3., 1.]])

    Args:
        column (int): Column in the (batched) matrix (..., M, N) over which to
            perform the summation
    """
    def __init__(self, column: int):
        super().__init__()
        warnings.warn('The Jacobian related function are not tested.', RuntimeWarning)
        self.column = column

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        y = x.clone()
        y[..., self.column] = y[..., self.column].cumsum(-1)
        return y

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        x = y.clone()
        x[..., self.column] = diff(x[..., self.column], -1)
        return x

    def log_det_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 1]:
        return torch.zeros_like(x[...,:1])

    def log_diag_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        return torch.zeros_like(x)

class DiffColumn(CumsumColumn):
    """
    Inverse of CumsumColumn transformation.
    """
    def forward(self, x, **kwargs):
        return super().inverse(x, **kwargs)

    def inverse(self, y, **kwargs):
        return super().forward(y, **kwargs)



from torchtyping import TensorType

import torch
from stribor import ElementwiseTransform


class Identity(ElementwiseTransform):
    """
    Identity transformation.
    Doesn't change the input, log-Jacobian is 0.
    """
    def __init__(self, **kwargs):
        super().__init__()

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return x

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return y

    def log_det_jacobian(self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 1]:
        return torch.zeros_like(x[...,:1])

    def log_diag_jacobian(self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return torch.zeros_like(x)



import pytest
import stribor as st
import torch

@pytest.mark.parametrize('input_dim', [(1, 1, 1), (10, 1, 1), (4, 4, 1), (3, 5, 2), (5, 3, 2, 3)])
@pytest.mark.parametrize('hidden_dims', [[], [13], [47, 11]])
@pytest.mark.parametrize('out_dim', [1, 2, 7])
@pytest.mark.parametrize('pooling', ['mean', 'max'])
def test_exclusive_set_net_zero_trace(input_dim, hidden_dims, out_dim, pooling):
    torch.manual_seed(123)

    dim = input_dim[-1]
    model = st.net.DiffeqZeroTraceDeepSet(dim, hidden_dims, dim * out_dim, pooling, return_log_det_jac=False)
    x = torch.randn(*input_dim).requires_grad_(True)

    t = torch.Tensor([1])
    f = lambda x: model(t, x)
    trace_exact = st.util.divergence_from_jacobian(f, x)

    assert trace_exact.sum() == 0
    assert torch.allclose(trace_exact, torch.zeros_like(trace_exact), atol=1e-6)


@pytest.mark.parametrize('input_dim', [(1, 1, 1), (10, 1, 1), (4, 4, 1), (3, 5, 2), (5, 3, 2, 3)])
@pytest.mark.parametrize('pooling', ['mean', 'max'])
def test_exclusive_set_net_masking(input_dim, pooling):
    torch.manual_seed(123)

    model = st.net.DiffeqZeroTraceDeepSet(input_dim[-1], [32], input_dim[-1] * 4, pooling, return_log_det_jac=False)

    # Regular masked output
    x = torch.randn(*input_dim)
    mask = torch.rand(*input_dim[:-1], 1).round()
    t = torch.Tensor([1])
    y1 = model(t, x, mask)

    # Output after changing masked elements
    x_perm = x + x * (1 - mask) * torch.rand(*input_dim)
    y2 = model(t, x_perm, mask)

    assert torch.isclose(y1, y2, atol=1e-5).all()



import torch
import pytest
import stribor as st
from stribor.test.base import *

@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (7, 4, 5), (2, 3, 4, 5)])
@pytest.mark.parametrize('dim', [-1])
def test_cumsum(input_shape, dim):
    torch.manual_seed(123)

    f = st.Cumsum(dim)
    x = torch.randn(*input_shape)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)

    y = f(x)
    assert torch.all(y == x.cumsum(dim))


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (7, 4, 5), (2, 3, 4, 5)])
@pytest.mark.parametrize('dim', [-1])
def test_diff(input_shape, dim):
    torch.manual_seed(123)

    f = st.Diff(dim)
    x = torch.randn(*input_shape)
    y = x.cumsum(dim)

    check_inverse_transform(f, y)
    check_log_jacobian_determinant(f, y)

    x_ = f(y)
    assert torch.allclose(x, x_)


@pytest.mark.parametrize('input_shape', [(2, 10), (7, 4, 5), (2, 3, 4, 5)])
@pytest.mark.parametrize('column', [1, -1, 3])
def test_cumsum_column(input_shape, column):
    torch.manual_seed(123)

    f = st.CumsumColumn(column)
    x = torch.randn(*input_shape)

    check_inverse_transform(f, x)



from .mlp import *
from .diffeq import *
from .diffeq_exact_trace import *
from .diffeq_zero_trace import *
from .diagjac import *
from .equivariant import *
from .attention import *
from .made import *
from .time_net import *
from .resnet import *



from .activations import *
from .affine import *
from .cnf import *
from .coupling import *
from .cumsum import *
from .identity import *
from .permute import *
from .iresnet import *
from .sigmoid import *
from .spline import *



from typing import List, Union
from torchtyping import TensorType
from abc import ABCMeta, abstractmethod

import torch
import torch.nn as nn
import stribor as st

#####################################################
# Use `divergence=approximate` in ContinuousTransform
#####################################################

class DiffeqNet(nn.Module, metaclass=ABCMeta):
    @abstractmethod
    def forward(
        self,
        t: TensorType[1],
        x: TensorType[..., 'dim'],
        latent: TensorType[..., 'latent'] = None,
        **kwargs,
    ) -> TensorType[..., 'out']:
        pass


class DiffeqConcat(DiffeqNet):
    """
    Differential equation that concatenates the input with time.

    Args:
        net (nn.Module): Neural network with `dim + 1 (+ latent)`
        input size and `dim` output size.
    """
    def __init__(self, net):
        super().__init__()
        self.net = net

    def forward(
        self,
        t: TensorType[1],
        x: TensorType[..., 'dim'],
        latent: TensorType[..., 'latent'] = None,
        **kwargs,
    ) -> TensorType[..., 'out']:
        t = torch.ones_like(x[..., :1]) * t
        input = torch.cat([t, x], -1)
        if latent is not None:
            input = torch.cat([input, latent], -1)
        return self.net(input, **kwargs)


class DiffeqMLP(DiffeqConcat):
    """
    Differential equation defined with MLP.

    Example:
    >>> batch, dim = 32, 3
    >>> net = stribor.net.DiffeqMLP(dim + 1, [64, 64], dim)
    >>> x = torch.randn(batch, dim)
    >>> t = torch.rand(batch, 1)
    >>> net(t, x).shape
    torch.Size([32, 3])

    Args:
        Same as in `st.net.MLP`
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        activation: str = 'Tanh',
        final_activation: str = None,
        **kwargs,
    ):
        super().__init__(st.net.MLP(in_dim, hidden_dims, out_dim, activation, final_activation))


class DiffeqDeepset(DiffeqConcat):
    """
    Differential equation defined with permutation equivariant network.

    Args:
        Same as in `st.net.EquivariantNet`
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        activation: str = 'Tanh',
        final_activation: str = None,
        **kwargs,
    ):
        super().__init__(st.net.EquivariantNet(in_dim, hidden_dims, out_dim, activation, final_activation))


class DiffeqSelfAttention(DiffeqConcat):
    """
    Differential equation defined with self attention network.

    Args:
        Same as in `st.net.SelfAttention`
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dim: List[int],
        out_dim: int,
        n_heads: int = 1,
        mask_diagonal: bool = False,
        **kwargs,
    ):
        super().__init__(st.net.SelfAttention(in_dim, hidden_dim, out_dim, n_heads, mask_diagonal))



from abc import ABCMeta, abstractmethod
from typing import List, Optional, Tuple, Union
from torchtyping import TensorType

import torch.nn as nn
from torch.distributions import Distribution

class Transform(nn.Module, metaclass=ABCMeta):
    @abstractmethod
    def forward(
        self, x: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        pass

    @abstractmethod
    def inverse(
        self, y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        pass

    @abstractmethod
    def log_det_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 1]:
        pass

    def jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim', 'dim']:
        raise NotImplementedError

    def forward_and_log_det_jacobian(
        self, x: TensorType[..., 'dim'], **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        y = self.forward(x, **kwargs)
        log_det_jacobian = self.log_det_jacobian(x, y, **kwargs)
        return y, log_det_jacobian

    def inverse_and_log_det_jacobian(
        self, y: TensorType[..., 'dim'], **kwargs
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        x = self.inverse(y, **kwargs)
        log_det_jacobian = self.log_det_jacobian(x, y, **kwargs)
        return x, -log_det_jacobian


class ElementwiseTransform(Transform):
    @abstractmethod
    def log_diag_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        pass

    def forward_and_log_diag_jacobian(
        self, x: TensorType[..., 'dim'], **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 'dim']]:
        y = self.forward(x, **kwargs)
        log_diag_jacobian = self.log_diag_jacobian(x, y, **kwargs)
        return y, log_diag_jacobian

    def inverse_and_log_diag_jacobian(
        self, y: TensorType[..., 'dim'], **kwargs
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 'dim']]:
        x = self.inverse(y, **kwargs)
        log_diag_jacobian = self.log_diag_jacobian(x, y, **kwargs)
        return x, -log_diag_jacobian


class NormalizingFlow(Transform):
    """
    Normalizing flow for density estimation and efficient sampling.

    Example:
    >>> import stribor as st
    >>> torch.manual_seed(123)
    >>> dim = 2
    >>> f = st.NormalizingFlow(st.UnitNormal(dim), [st.Affine(dim)])
    >>> f.log_prob(torch.randn(3, 2))
    tensor([[-1.7560], [-1.7434], [-2.1792]])
    >>> f.sample(2)
    tensor([[-0.5204,  0.4196]])

    Args:
        base_dist (torch.distributions.Distribution): Base distribution
        transforms (Transform): List of invertible transformations
    """
    def __init__(
        self,
        base_dist: Distribution,
        transforms: List[Transform],
    ):
        super().__init__()
        self.base_dist = base_dist
        self.transforms = nn.ModuleList(transforms)

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        for f in self.transforms:
            x = f(x, **kwargs)
        return x

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        for f in reversed(self.transforms):
            y = f.inverse(y, **kwargs)
        return y

    def forward_and_log_det_jacobian(
        self, x: TensorType[..., 'dim'], **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        log_det_jac = 0
        for f in self.transforms:
            x, ldj = f.forward_and_log_det_jacobian(x, **kwargs)
            log_det_jac += ldj
        return x, log_det_jac

    def inverse_and_log_det_jacobian(
        self, y: TensorType[..., 'dim'], **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        log_det_jac = 0
        for f in reversed(self.transforms):
            y, ldj = f.inverse_and_log_det_jacobian(y, **kwargs)
            log_det_jac += ldj
        return y, log_det_jac

    def log_prob(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 1]:
        x, log_det_jac = self.inverse_and_log_det_jacobian(y, **kwargs)
        log_prob = self.base_dist.log_prob(x).unsqueeze(-1) + log_det_jac
        return log_prob

    def sample(
        self, num_samples: Union[Tuple[int], int], *, rsample: bool = False, **kwargs,
    ) -> TensorType:
        if isinstance(num_samples, int):
            num_samples = (num_samples,)

        if rsample:
            x = self.base_dist.rsample(num_samples)
        else:
            x = self.base_dist.sample(num_samples)
        x = self.forward(x, **kwargs)
        return x

    def rsample(self, num_samples: Union[Tuple[int], int], **kwargs) -> TensorType:
        return self.sample(num_samples, **kwargs)

    def log_det_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 1]:
        _, log_det_jacobian = self.forward_and_log_det_jacobian(x, **kwargs)
        return log_det_jacobian


class NeuralFlow(nn.Module):
    """
    Neural flow model.
    https://arxiv.org/abs/2110.13040

    Example:
    >>> import stribor as st
    >>>

    Args:
        transforms (Transform): List of invertible transformations
            that satisfy initial condition F(x, t=0)=x.
    """
    def __init__(self, transforms: List[Transform]) -> None:
        super().__init__()
        self.transforms = nn.ModuleList(transforms)

    def forward(
        self,
        x: TensorType[..., 'dim'],
        t: TensorType[..., 1],
        t0: Optional[TensorType[..., 1]] = None,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        if t0 is not None:
            for transform in reversed(self.transforms):
                x = transform.inverse(x, t=t0, **kwargs)
        for transform in self.transforms:
            x = transform(x, t=t, **kwargs)
        return x



from typing import Union
from torchtyping import TensorType
from numbers import Number

import torch
import torch.distributions as td

class Normal(td.Independent):
    """
    Normal distribution on the d-dimensional space defined by
    mean and std tensors of any shape.

    Example:
    >>> dist = stribor.Normal(0., 1.)
    >>> dist.log_prob(torch.Tensor([0]))
    tensor([-0.9189])
    >>> dist = stribor.Normal(torch.zeros(2), torch.ones(2))
    >>> dist.log_prob(torch.zeros(2, 2))
    tensor([-1.8379, -1.8379])

    Args:
        loc (float or tensor): Mean
        scale (float or tensor): Standard deviation
    """
    def __init__(
        self,
        loc: Union[Number, TensorType[...]],
        scale: Union[Number, TensorType[...]],
        **kwargs,
    ):
        self.loc = loc
        self.scale = scale

        # Support float input
        rbd = 0 if isinstance(self.loc, float) else 1

        super().__init__(td.Normal(self.loc, self.scale, **kwargs), reinterpreted_batch_ndims=rbd)


class UnitNormal(Normal):
    """
    Unit `st.Normal`. Specify only the dimension.

    Example:
    >>> dist = stribor.UnitNormal(2)
    >>> dist.log_prob(torch.ones(1, 2))
    tensor([-2.8379])

    Args:
        dim (int): Dimension of data
    """
    def __init__(self, dim: int, **kwargs):
        self.dim = dim
        super().__init__(torch.zeros(self.dim), torch.ones(self.dim), **kwargs)


class MultivariateNormal(td.MultivariateNormal):
    """
    Wrapper for `torch.distributions.MultivariateNormal`.

    Args:
        loc (Tensor): mean of the distribution
        covariance_matrix (Tensor): positive-definite covariance matrix
        precision_matrix (Tensor): positive-definite precision matrix
        scale_tril (Tensor): lower-triangular factor of covariance, with positive-valued diagonal
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import pytest
import stribor as st
import torch

@pytest.mark.parametrize('dim', [2, 3, 11])
@pytest.mark.parametrize('latent_dim', [1, 3, 8])
def test_mlp_zero_trace(dim, latent_dim):
    torch.manual_seed(123)

    out_dim = dim * latent_dim
    model = st.net.DiffeqZeroTraceMLP(dim, [64, 64], out_dim)

    t = torch.Tensor([1])
    x = torch.randn(10, dim)
    y, jac = model(t, x)

    assert y.shape[:-1] == x.shape[:-1] and y.shape[-1] == out_dim

    f = lambda t, x: model(t, x)[0]
    jac_exact = st.util.divergence_from_jacobian(f, (t, x))[1]

    assert jac_exact.sum() == 0
    assert torch.isclose(jac_exact, jac).all()

@pytest.mark.parametrize('dim', [2, 3, 11])
@pytest.mark.parametrize('latent_dim', [1, 3, 8])
@pytest.mark.parametrize('pooling', ['max', 'sum', 'mean'])
def test_deepset_zero_trace(dim, latent_dim, pooling):
    torch.manual_seed(123)

    out_dim = dim * latent_dim
    model = st.net.DiffeqZeroTraceDeepSet(dim, [64, 64], out_dim, pooling=pooling)

    t = torch.Tensor([1])
    x = torch.randn(5, 10, dim)

    y, jac = model(t, x)

    assert len(y) == len(x) and y.shape[:-1] == x.shape[:-1] and y.shape[-1] == out_dim

    f = lambda t, x: model(t, x)[0]
    jac_exact = st.util.divergence_from_jacobian(f, (t, x))[1]

    assert jac_exact.sum() == 0
    assert torch.isclose(jac_exact, jac).all()

@pytest.mark.parametrize('input_dim', [(2, 5, 7), (3, 4, 5, 2)])
@pytest.mark.parametrize('latent_dim', [1, 8])
@pytest.mark.parametrize('n_heads', [1, 2])
def test_attention_zero_trace(input_dim, latent_dim, n_heads):
    torch.manual_seed(123)

    dim = input_dim[-1]
    model = st.net.DiffeqZeroTraceAttention(dim, [64, 64], dim * latent_dim, n_heads=n_heads)

    t = torch.Tensor([1])
    x = torch.randn(*input_dim)

    y, div = model(t, x)
    f = lambda t, x: model(t, x)[0]
    _, div_exact = st.util.divergence_from_jacobian(f, (t, x))

    assert div_exact.sum() == 0
    assert torch.allclose(div, div_exact)



from itertools import chain
import torch

__all__ = ['flatten_params']

def _flatten(sequence):
    flat = [p.contiguous().view(-1) for p in sequence]
    return torch.cat(flat) if len(flat) > 0 else torch.tensor([])

def flatten_params(*nets):
    return _flatten(chain(*[x.parameters() for x in nets]))



import torch
import pytest
import numpy as np
import stribor as st
from stribor.test.base import *


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('n_bins', [1, 3, 10])
@pytest.mark.parametrize('latent_dim', [0, 1, 13])
@pytest.mark.parametrize('spline', ['cubic', 'quadratic'])
def test_spline(input_shape, n_bins, latent_dim, spline):
    np.random.seed(123)
    torch.manual_seed(123)

    x = torch.rand(*input_shape) * 2
    latent = torch.randn(*input_shape[:-1], latent_dim) if latent_dim != 0 else None

    dim = input_shape[-1]
    out_spline_dim = dim * (2 * n_bins + 2) if spline == 'cubic' else dim * (3 * n_bins - 1)

    f = st.Spline(
        dim=dim,
        n_bins=n_bins,
        latent_net=st.net.MLP(latent_dim, [12], out_spline_dim) if latent is not None else None,
        lower=0,
        upper=2,
        spline_type=spline,
    )

    check_inverse_transform(f, x, latent=latent)
    check_log_jacobian_determinant(f, x, latent=latent)
    check_gradients_not_nan(f, x, latent=latent)


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('spline', ['quadratic'])
def test_quadratic_spline_free_bounding_box_param(input_shape, spline):
    torch.manual_seed(123)

    x = torch.rand(*input_shape)

    class UnboundedSpline(st.Spline):
        def forward_and_log_diag_jacobian(self, x, latent=None, *, reverse=False, **kwargs):
            w, h, d = self._get_params(latent)
            return self.spline(x, w, h, d, inverse=reverse, left=-1, bottom=-3, top=2, right=1)

    f = UnboundedSpline(dim=input_shape[-1], n_bins=5, spline_type=spline)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)
    check_gradients_not_nan(f, x)



import pytest
import stribor as st
import torch

@pytest.mark.parametrize('input_dim', [(1, 1, 1), (10, 3, 2), (5, 3, 2, 3), (3, 2, 4, 6, 7)])
@pytest.mark.parametrize('hidden_dims', [[32], [64, 32]])
@pytest.mark.parametrize('out_dim', [1, 2, 5])
@pytest.mark.parametrize('n_heads', [1, 4, 8])
@pytest.mark.parametrize('mask_diagonal', [True, False])
@pytest.mark.parametrize('model_name', ['Attention', 'SelfAttention', 'InducedSelfAttention'])
def test_attention_equivariance_and_masking(input_dim, hidden_dims, out_dim, n_heads, mask_diagonal, model_name):
    torch.manual_seed(123)

    model = getattr(st.net, model_name)(input_dim[-1], hidden_dims, out_dim,
        n_heads=n_heads, mask_diagonal=mask_diagonal, n_points=11)

    x = torch.randn(*input_dim)
    if model_name == 'Attention':
        y = model(x, x, x)
    else:
        y = model(x)

    # Check if permutation equivariant
    x_flip = torch.flip(x, [-2])
    if model_name == 'Attention':
        y_flip = model(x_flip, x_flip, x_flip)
    else:
        y_flip = model(x_flip)
    y_flip = torch.flip(y_flip, [-2]) # Flip back to original order

    assert torch.isclose(y, y_flip, atol=1e-5).all(), 'Attention should be permutation invariant'

    # Check if masking works
    mask = torch.rand(*input_dim[:-1], 1).round()
    mask[...,0,0] = 1 # Make sure at least one element is not masked

    x_perm = x + x * (1 - mask) * torch.rand(*input_dim)

    if model_name == 'Attention':
        y = model(x, x, x, mask)
        y_perm = model(x_perm, x_perm, x_perm, mask)
    else:
        y = model(x, mask)
        y_perm = model(x_perm, mask)

    assert torch.isclose(y, y_perm, atol=1e-5).all(), 'Masking in attention doesn\'t work'



from typing import Union
from torchtyping import TensorType
from numbers import Number

import torch
import torch.distributions as td

class Uniform(td.Independent):
    """
    Uniform distribution on the d-dimensional space defined by
    low and high tensors of any shape.

    Example:
    >>> dist = stribor.Uniform(0., 1.)
    >>> dist.log_prob(torch.zeros(1, 2))
    tensor([[0., 0.]])
    >>> dist = stribor.Uniform(torch.zeros(2), torch.ones(2))
    >>> dist.log_prob(torch.zeros(1, 2))
    tensor([0.])

    Args:
        low (float or tensor): Lower bound (inclusive)
        high (float or tensor): Upper bound (exclusive)
    """
    def __init__(
        self,
        low: Union[Number, TensorType[...]],
        high: Union[Number, TensorType[...]],
        **kwargs,
    ):
        self.low = low
        self.high = high

        rbd = 0 if isinstance(self.low, float) else 1

        super().__init__(td.Uniform(self.low, self.high, **kwargs), reinterpreted_batch_ndims=rbd)

class UnitUniform(Uniform):
    """
    Unit `st.Uniform`, i.e., on interval [0, 1). Specify only the dimension.

    Example:
    >>> dist = stribor.UnitUniform(2)
    >>> dist.log_prob(torch.zeros(1, 2))
    tensor([0.])

    Args:
        dim (int): Dimension of data
    """
    def __init__(self, dim: int):
        self.dim = dim
        super().__init__(torch.zeros(self.dim), torch.ones(self.dim))



from typing import Union, List
from torchtyping import TensorType

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['MADE']

class MaskedLinear(nn.Linear):
    """
    Same as torch.nn.Linear except has a mask on the weights.

    Args:
        in_features (int): Input data dimension
        out_features (int): Output data dimension
        bias (bool): Whether to use bias. Default: True
    """
    def __init__(self, in_features: int, out_features: int, bias=True):
        super().__init__(in_features, out_features, bias)
        self.register_buffer('mask', torch.ones(out_features, in_features))

    def set_mask(self, mask):
        self.mask.data.copy_(torch.from_numpy(mask.astype(np.uint8).T))

    def forward(self, input):
        return F.linear(input, self.mask * self.weight, self.bias)

class MADE(nn.Module):
    """
    Autoregressive transformation with a neural network, implemented using weight masking.
    "MADE: Masked Autoencoder for Distribution Estimation" (https://arxiv.org/abs/1502.03509)

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size. Has to be multiple of `in_dim` such that each dimension
            has assigned output dimensions. `return_per_dim` governs how the output looks like
        activation (str): Activation function from `torch.nn`. Default: 'Tanh'
        final_activation (str): Last activation. Default: None
        num_masks (int): Number of ordering ensembles. Default: 1
        natural_ordering (bool): Whether to use natural order, else random. Default: False
        reverse_ordering (bool): Whether to reverse the order. Default: False
        return_per_dim (bool): Whether to return in (..., in_dim, out_dim / in_dim) format,
            otherwise returns (..., out_dim). Default: False
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        activation: str = 'Tanh',
        final_activation: str = None,
        num_masks: int = 1,
        natural_ordering: bool = False,
        reverse_ordering: bool = False,
        return_per_dim: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.hidden_dims = hidden_dims
        self.activation = activation
        self.final_activation = final_activation
        self.return_per_dim = return_per_dim
        self.natural_ordering = natural_ordering
        self.reverse_ordering = reverse_ordering
        self.num_masks = num_masks

        assert self.out_dim % self.in_dim == 0, "out_dim must be integer multiple of in_dim"

        self.net = self._get_net()

        self.m = {}
        self.update_masks() # builds the initial self.m connectivity

    def _get_net(self):
        net = []
        hs = [self.in_dim] + self.hidden_dims + [self.out_dim]
        for h0, h1 in zip(hs, hs[1:]):
            net.append(MaskedLinear(h0, h1))
            net.append(getattr(nn, self.activation)())
        net.pop() # pop the last activation for the output layer
        if self.final_activation is not None:
            net.append(getattr(nn, self.final_activation)())
        return nn.Sequential(*net)

    def update_masks(self):
        if self.m and self.num_masks == 1:
            return # only a single seed, skip for efficiency

        L = len(self.hidden_dims)

        rng = np.random.RandomState()

        # sample the order of the inputs and the connectivity of all neurons
        if self.natural_ordering:
            self.m[-1] = np.arange(self.in_dim)
            if self.reverse_ordering:
                self.m[-1] = self.m[-1][::-1]
        else:
            self.m[-1] = rng.permutation(self.in_dim)

        for l in range(L):
            self.m[l] = rng.randint(self.m[l-1].min(), max(self.in_dim-1, 1), size=self.hidden_dims[l])

        # construct the mask matrices
        masks = [self.m[l-1][:,None] <= self.m[l][None,:] for l in range(L)]
        masks.append(self.m[L-1][:,None] < self.m[-1][None,:])

        # handle the case where out_dim = in_dim * k, for integer k > 1
        if self.out_dim > self.in_dim:
            k = int(self.out_dim / self.in_dim)
            # replicate the mask across the other outputs
            masks[-1] = np.concatenate([masks[-1]]*k, axis=1)

        # set the masks in all MaskedLinear layers
        layers = [l for l in self.net.modules() if isinstance(l, MaskedLinear)]
        for l, m in zip(layers, masks):
            l.set_mask(m)

    def forward(
        self, x: TensorType[..., 'dim'], **kwargs,
    ) -> Union[TensorType[..., 'out'], TensorType[..., 'dim', 'out/dim']]:
        original_shape = x.shape

        x = x.view(-1, original_shape[-1])
        y = self.net(x)

        y = y.view(*original_shape[:-1], -1, self.in_dim).transpose(-1, -2)

        return y if self.return_per_dim else y.reshape(*y.shape[:-2], -1)



import pytest
import torch
import stribor as st


@pytest.mark.parametrize('input_dim', [(1, 1), (2, 3, 1)])
@pytest.mark.parametrize('output_dim', [1, 5])
@pytest.mark.parametrize('function', ['TimeIdentity', 'TimeLinear', 'TimeTanh', 'TimeLog', 'TimeFourier', 'TimeFourierBounded'])
def test_time_derivative(input_dim, output_dim, function):
    f = getattr(st.net.time_net, function)(output_dim, hidden_dim=8)

    t = torch.rand(*input_dim) * 5
    output = f(t)
    dt = f.derivative(t)

    assert dt.shape == output.shape and output.shape[-1] == output_dim
    assert len(t.shape) == len(output.shape) and t.shape[:-1] == output.shape[:-1]

    dt_true = torch.autograd.functional.jacobian(f, t, vectorize=True)
    for _ in range(len(dt_true.shape) // 2):
        dt_true = dt_true.sum(-1)

    assert torch.allclose(dt, dt_true, atol=1e-4)



import torch
import stribor as st
from stribor.test.base import *

def test_normalizing_flow():
    torch.manual_seed(123)

    dim = 2
    n_bins = 3
    hidden_dims = [13]

    f = st.NormalizingFlow(
        st.UnitNormal(dim),
        [
            st.Coupling(st.Affine(dim, latent_net=st.net.MLP(dim, hidden_dims, 2 * dim)), mask='ordered_1'),
            st.ContinuousTransform(
                dim,
                net=st.net.DiffeqMLP(dim + 1, hidden_dims, dim),
                divergence='compute',
                atol=1e-8,
                rtol=1e-8,
            ),
            st.Flip(dims=[-1]),
            st.Sigmoid(),
            st.Coupling(
                st.Spline(
                    dim,
                    n_bins=n_bins,
                    latent_net=st.net.MLP(dim, hidden_dims, st.util.cubic_spline_latent_dim(dim, n_bins)),
                    spline_type='cubic',
                ),
                mask='ordered_0',
            ),
            st.Logit(),
        ]
    )

    x = torch.randn(3, 4, dim)
    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)
    check_gradients_not_nan(f, x)
    check_area_under_pdf_2D(f)


def test_doc_example():
    torch.manual_seed(123)
    dim = 2

    f = st.NormalizingFlow(st.UnitNormal(dim), [st.Affine(dim)])

    lp = f.log_prob(torch.randn(3, 2))
    assert torch.allclose(lp, torch.Tensor([[-1.7560], [-1.7434], [-2.1792]]), atol=1e-4)

    s = f.sample(1)
    assert torch.allclose(s, torch.Tensor([[-0.5204,  0.4196]]), atol=1e-4)



from .mask import *
from .search_sorted import *
from .rational_quadratic_spline import *
from .cubic_spline import *
from .divergence import *
from .flatten_params import *
from .safe_softmax import *



import stribor as st
import torch
from torch.nn import functional as F
import numpy as np


def quadratic_spline_latent_dim(dim: int, n_bins) -> int:
    return dim * (3 * n_bins - 1)


def unconstrained_rational_quadratic_spline(inputs,
                                            unnorm_widths,
                                            unnorm_heights,
                                            unnorm_derivatives,
                                            inverse=False,
                                            lower=-1.,
                                            upper=1.,
                                            left=None,
                                            right=None,
                                            bottom=None,
                                            top=None,
                                            min_bin_width=1e-3,
                                            min_bin_height=1e-3,
                                            min_derivative=1e-3):
    """
    Takes inputs and unnormalized parameters for width, height and
    derivatives of spline bins. Normalizes parameters and applies quadratic spline.
    The domain and codomain can be defined with (lower, upper) or the domain can be
    defined with (left, right) and codomain with (bottom, top).
    Code adapted from https://github.com/bayesiains/nsf

    Args:
        inputs (tensor): Input with shape (..., dim)
        unnorm_widths (tensor): Bin widths (x-axis) with shape (..., dim, n_bins)
        unnorm_heights (tensor): Bin heights (y-axis) between knots with shape (..., dim, n_bins)
        unnorm_derivatives (tensor): Derivatives in knots with shape (..., dim, n_bins - 1) or
            (..., dim, n_bins + 1) (whether derivative in bounds is specified)
        inverse (bool, optional): Whether to invert the calculation. Default: False
        lower (float, optional): Lower domain/codomain bound. Default: -1.
        upper (float, optional): Upper domain/codomain bound. Default: 1.
        left (float or tensor, optional): Lower domain bound. Default: None
        right (float or tensor, optional): Upper domain bound. Default: None
        bottom (float or tensor, optional): Lower codomain bound. Default: None
        top (float or tensor, optional): Upper codomain bound. Default: None
        min_bin_width (float, optional): Minimum bin width. Default: 1e-3
        min_bin_height (float, optional): Minimum bin height. Default: 1e-3
        min_derivative (float, optional): Minimum knot derivative. Default: 1e-3

    Returns:
        outputs (tensor): Spline transformed input (..., dim)
        ljd (tensor): Log-Jacobian diagonal (..., dim)
    """

    # Check if all boundaries are defined
    if all(x is not None for x in [left, right, top, bottom]):
        if inverse:
            lower = bottom
            upper = top
        else:
            lower = left
            upper = right
    else:
        left = bottom = lower
        right = top = upper

    # Define inside/outside spline window
    unnorm_widths = unnorm_widths.expand(*inputs.shape, -1)
    unnorm_heights = unnorm_heights.expand(*inputs.shape, -1)
    unnorm_derivatives = unnorm_derivatives.expand(*inputs.shape, -1)

    inside_interval = (inputs >= lower) & (inputs <= upper)
    outside_interval = ~inside_interval

    # Define outputs
    outputs = torch.zeros_like(inputs)
    ljd = torch.zeros_like(inputs)

    # If edge derivatives are not parametrized, set to constant
    if unnorm_derivatives.shape[-1] == unnorm_widths.shape[-1] - 1:
        unnorm_derivatives = F.pad(unnorm_derivatives, pad=(1, 1))
        constant = np.log(np.exp(1 - min_derivative) - 1)
        unnorm_derivatives[..., 0] = constant
        unnorm_derivatives[..., -1] = constant

    # Define linear tails (outside domain)
    outputs[outside_interval] = inputs[outside_interval]
    ljd[outside_interval] = 0

    # If no points are inside domain -> return unchanged
    if not inside_interval.any():
        return outputs, ljd

    # Go from unconstrained to actual values
    num_bins = unnorm_widths.shape[-1]

    if min_bin_width * num_bins > 1.0:
        raise ValueError('Minimal bin width too large for the number of bins')
    if min_bin_height * num_bins > 1.0:
        raise ValueError('Minimal bin height too large for the number of bins')

    widths = F.softmax(unnorm_widths, dim=-1)
    widths = min_bin_width + (1 - min_bin_width * num_bins) * widths

    heights = F.softmax(unnorm_heights, dim=-1)
    heights = min_bin_height + (1 - min_bin_height * num_bins) * heights

    derivatives = min_derivative + F.softplus(unnorm_derivatives)

    # Rational spline
    outputs, ljd = rational_quadratic_spline(
        inputs=inputs,
        widths=widths,
        heights=heights,
        derivatives=derivatives,
        inside_interval=inside_interval,
        initial_outputs=outputs,
        initial_ljd=ljd,
        inverse=inverse,
        left=left,
        right=right,
        bottom=bottom,
        top=top
    )

    return outputs, ljd


def rational_quadratic_spline(inputs,
                              widths,
                              heights,
                              derivatives,
                              inside_interval,
                              initial_outputs,
                              initial_ljd,
                              left,
                              right,
                              bottom,
                              top,
                              inverse=False):
    """
    Args:
        inputs (tensor): Input with shape (..., dim)
        widths (tensor): Bin widths with shape (..., dim, n_bins)
        heights (tensor): Bin heights with shape (..., dim, n_bins)
        derivatives (tensor): Derivatives with shape (..., dim, n_bins + 1)
        inside_interval (tensor): Boolean mask, (..., dim)
        initial_outputs: Initial input, usually initialized to zero with shape (..., dim)
        initial_ljd: Same as initial_outputs
        left (float or tensor): Left boundary, if tensor, has shape (..., dim)
        right (float or tensor): Right boundary, same type as left boundary
        bottom (float or tensor): Bottom boundary, same type as left boundary
        top (float or tensor): Top boundary, same type as left boundary
        inverse (bool, optional): Whether to do inverse calculation. Default: False

    Returns:
        outputs: (..., dim)
        ljd: (..., dim)
    """

    # Take only values inside interval
    inputs = inputs[inside_interval]
    widths = widths[inside_interval, :]
    heights = heights[inside_interval, :]
    derivatives = derivatives[inside_interval, :]

    # If boundaries are not tensors, convert them to tensors
    def boundary_to_tensor(b):
        return b[inside_interval].unsqueeze(-1) if torch.is_tensor(b) else torch.ones(inputs.shape[0], 1) * b
    left = boundary_to_tensor(left)
    right = boundary_to_tensor(right)
    top = boundary_to_tensor(top)
    bottom = boundary_to_tensor(bottom)

    # Check if values fall out of domain
    if inverse and ((inputs < bottom).any() or (inputs > top).any()):
        raise ValueError('Inverse input is outside of domain')
    elif not inverse and ((inputs < left).any() or (inputs > right).any()):
        raise ValueError('Input is outside of domain')

    cumwidths = torch.cumsum(widths, dim=-1)
    cumwidths = F.pad(cumwidths, pad=(1, 0), mode='constant', value=0.0)
    cumwidths = (right - left) * cumwidths + left
    cumwidths[..., 0, None] = left
    cumwidths[..., -1, None] = right
    widths = cumwidths[..., 1:] - cumwidths[..., :-1]

    cumheights = torch.cumsum(heights, dim=-1)
    cumheights = F.pad(cumheights, pad=(1, 0), mode='constant', value=0.0)
    cumheights = (top - bottom) * cumheights + bottom
    cumheights[..., 0, None] = bottom
    cumheights[..., -1, None] = top
    heights = cumheights[..., 1:] - cumheights[..., :-1]

    if inverse:
        bin_idx = st.util.searchsorted(cumheights, inputs)[..., None]
    else:
        bin_idx = st.util.searchsorted(cumwidths, inputs)[..., None]

    input_cumwidths = cumwidths.gather(-1, bin_idx)[..., 0]
    input_bin_widths = widths.gather(-1, bin_idx)[..., 0]

    input_cumheights = cumheights.gather(-1, bin_idx)[..., 0]
    delta = heights / widths
    input_delta = delta.gather(-1, bin_idx)[..., 0]

    input_derivatives = derivatives.gather(-1, bin_idx)[..., 0]
    input_derivatives_plus_one = derivatives[..., 1:].gather(-1, bin_idx)[..., 0]

    input_heights = heights.gather(-1, bin_idx)[..., 0]

    if inverse:
        a = (((inputs - input_cumheights) * (input_derivatives
                                             + input_derivatives_plus_one
                                             - 2 * input_delta)
              + input_heights * (input_delta - input_derivatives)))
        b = (input_heights * input_derivatives
             - (inputs - input_cumheights) * (input_derivatives
                                              + input_derivatives_plus_one
                                              - 2 * input_delta))
        c = - input_delta * (inputs - input_cumheights)

        discriminant = b.pow(2) - 4 * a * c
        assert (discriminant >= 0).all()

        root = (2 * c) / (-b - torch.sqrt(discriminant))
        outputs = root * input_bin_widths + input_cumwidths

        theta_one_minus_theta = root * (1 - root)
        denominator = input_delta + ((input_derivatives + input_derivatives_plus_one - 2 * input_delta)
                                     * theta_one_minus_theta)
        derivative_numerator = input_delta.pow(2) * (input_derivatives_plus_one * root.pow(2)
                                                     + 2 * input_delta * theta_one_minus_theta
                                                     + input_derivatives * (1 - root).pow(2))
        ljd = -torch.log(derivative_numerator) + 2 * torch.log(denominator) # Note the sign change
    else:
        theta = (inputs - input_cumwidths) / input_bin_widths
        theta_one_minus_theta = theta * (1 - theta)

        numerator = input_heights * (input_delta * theta.pow(2)
                                     + input_derivatives * theta_one_minus_theta)
        denominator = input_delta + ((input_derivatives + input_derivatives_plus_one - 2 * input_delta)
                                     * theta_one_minus_theta)
        outputs = input_cumheights + numerator / denominator

        derivative_numerator = input_delta.pow(2) * (input_derivatives_plus_one * theta.pow(2)
                                                     + 2 * input_delta * theta_one_minus_theta
                                                     + input_derivatives * (1 - theta).pow(2))
        ljd = torch.log(derivative_numerator) - 2 * torch.log(denominator)

    initial_outputs[inside_interval], initial_ljd[inside_interval] = outputs, ljd
    return initial_outputs, initial_ljd



import torch

def safe_softmax(x, dim=-1):
    """
    Same as `torch.softmax` but returns 0 instead of nan when the whole row is -inf.
    The consequence is that not every row is guaranteed to sum up to one.

    Args:
        x (tensor): Input
        dim (int): Along which dim to perform softmax
    """
    x = torch.softmax(x, dim)
    x = torch.nan_to_num(x)
    return x



import numpy as np
import torch

__all__ = ['get_mask']

def get_mask(mask):
    if mask == 'none':
        return none()
    elif mask == 'ordered_right_half' or mask == 'ordered_0':
        return ordered(ratio_zero=0.5, right_zero=False)
    elif mask == 'ordered_left_half' or mask == 'ordered_1':
        return ordered(ratio_zero=0.5, right_zero=True)
    elif mask == 'random_half':
        return random(ratio_zero=0.5)
    elif mask == 'parity_even':
        return parity(even_zero=False)
    elif mask == 'parity_odd':
        return parity(even_zero=True)
    else:
        raise NotImplementedError()

def none():
    return lambda _: torch.Tensor([0])

def random(ratio_zero=0.5):
    def mask(dim):
        if dim == 1:
            return torch.Tensor([1])
        mask = np.zeros(dim)
        size = np.clip(int(dim * ratio_zero), 1, dim - 1)
        mask[np.random.choice(np.arange(dim), size, replace=False)] = 1
        return torch.Tensor(mask)
    return mask

def ordered(ratio_zero=0.5, right_zero=False):
    def mask(dim):
        if dim == 1:
            return torch.Tensor([1])
        mask = np.ones(dim)
        size = np.clip(int(dim * ratio_zero), 1, dim - 1)
        mask[:size] = 0
        if right_zero:
            mask = 1 - mask
        return torch.Tensor(mask)
    return mask

def parity(even_zero=False):
    # If even zero, first element (ind=0) will be 0, and so on
    def mask(dim):
        if dim == 1:
            return torch.Tensor([1])
        mask = np.ones(dim)
        mask[::2] = 0
        if even_zero:
            mask = 1 - mask
        return torch.Tensor(mask)
    return mask



import torch
import stribor as st

def test_neural_flow():
    torch.manual_seed(123)

    dim = 2

    f = st.NeuralFlow([
        st.ContinuousAffineCoupling(
            latent_net=st.net.MLP(dim, [32], 2 * dim),
            time_net=st.net.TimeLinear(dim),
            mask='ordered_0',
            concatenate_time=False,
        ),
        st.ContinuousIResNet(
            dim,
            [32, 32],
            time_net=st.net.TimeTanh(dim),
        )
    ])

    x = torch.randn(10, 4, 2)
    t = torch.zeros_like(x[...,:1])

    y = f(x, t=t)
    assert (x == y).all()

    t0 = torch.randn_like(x[...,:1])

    y = f(x, t=t0, t0=t0)
    assert torch.allclose(x, y)



from torchtyping import TensorType

import torch
import torch.nn.functional as F

from stribor import ElementwiseTransform


class Sigmoid(ElementwiseTransform):
    """
    Sigmoid transformation.

    Code adapted from torch.distributions.transforms
    """
    def __init__(self, **kwargs):
        super().__init__()

    def forward(
        self, x: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        finfo = torch.finfo(x.dtype)
        y = torch.clamp(torch.sigmoid(x), min=finfo.tiny, max=1. - finfo.eps)
        return y

    def inverse(
        self, y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        finfo = torch.finfo(y.dtype)
        y = y.clamp(min=finfo.tiny, max=1.0 - finfo.eps)
        x = y.log() - (-y).log1p()
        return x

    def log_det_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 1]:
        return self.log_diag_jacobian(x, y, **kwargs).sum(-1, keepdim=True)

    def log_diag_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        return -F.softplus(-x) - F.softplus(x)

class Logit(Sigmoid):
    """
    Logit transformation. Inverse of sigmoid.
    """
    def forward(self, x, **kwargs):
        return super().inverse(x, **kwargs)

    def inverse(self, y, **kwargs):
        return super().forward(y, **kwargs)

    def log_diag_jacobian(
        self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs,
    ) -> TensorType[..., 'dim']:
        return -super().log_diag_jacobian(y, x) # Flip sign and order



from typing import Union
from torchtyping import TensorType

import torch
from stribor import Transform, ElementwiseTransform


def check_inverse_transform(f: Transform, x: TensorType[..., 'dim'], **kwargs):
    y = f(x, **kwargs)
    x_ = f.inverse(y, **kwargs)
    assert torch.allclose(x, x_, atol=1e-4), 'Inverse function for a given forward function is not correct.'


def _check_log_det_operations(f, x, **kwargs):
    y = f(x, **kwargs)
    ljd = f.log_det_jacobian(x, y, **kwargs)

    _, ljd1 = f.forward_and_log_det_jacobian(x, **kwargs)
    assert torch.allclose(ljd, ljd1), 'Error in implementing `forward_and_log_diag_jacobian`'

    _, ljd2 = f.inverse_and_log_det_jacobian(y, **kwargs)
    assert torch.allclose(ljd, -ljd2, atol=1e-4), 'Error in implementing `inverse_and_log_diag_jacobian`'

def _get_full_jacobian(f, x, **kwargs):
    x = x.view(-1, x.shape[-1])
    kwargs = { k: v.view(-1, v.shape[-1]) if isinstance(v, torch.Tensor) else v for k,v in kwargs.items() }
    y = f(x, **kwargs)

    jacobian = torch.autograd.functional.jacobian(lambda v: f(v, **kwargs), (x,), strict=True)[0]
    jacobian = jacobian.permute(0, 2, 1, 3).sum(0)

    jacobian_inverse = torch.autograd.functional.jacobian(lambda v: f.inverse(v, **kwargs), (y,), strict=True)[0]
    jacobian_inverse = jacobian_inverse.permute(0, 2, 1, 3).sum(0)

    return x, kwargs, jacobian, jacobian_inverse

def _check_log_det_jacobian(f, x, jacobian, reverse=False, **kwargs):
    y = f(x, **kwargs)
    log_det_jacobian = torch.det(jacobian).abs().log()
    if reverse:
        _, log_det_jacobian_model = f.inverse_and_log_det_jacobian(y, **kwargs)
    else:
        _, log_det_jacobian_model = f.forward_and_log_det_jacobian(x, **kwargs)
    assert torch.allclose(log_det_jacobian, log_det_jacobian_model.squeeze(-1), atol=1e-4), 'Jacobian determinant is incorrect'

def _check_whole_jacobian(f, x, jacobian, **kwargs):
    y = f(x, **kwargs)
    try:
        jacobian_model = f.jacobian(x, y, **kwargs)
    except NotImplementedError:
        pass
    else:
        assert torch.allclose(jacobian_model, jacobian, atol=1e-4), 'Jacobian is incorrect'

def _check_log_diag_jacobian(f, x, jacobian, **kwargs):
    y = f(x, **kwargs)
    try:
        log_diag_jacobian_model = f.log_diag_jacobian(x, y, **kwargs)
    except AttributeError:
        pass
    else:
        log_diag_jacobian = torch.diagonal(jacobian, dim1=-2, dim2=-1).log()
        assert torch.allclose(log_diag_jacobian, log_diag_jacobian_model, atol=1e-4), 'Jacobian diagonal is incorrect'

def check_log_jacobian_determinant(f: Union[Transform, ElementwiseTransform], x: TensorType[..., 'dim'], **kwargs):
    _check_log_det_operations(f, x, **kwargs)

    x, kwargs, jacobian, jacobian_inverse = _get_full_jacobian(f, x, **kwargs)

    _check_log_det_jacobian(f, x, jacobian, **kwargs)
    _check_log_det_jacobian(f, x, jacobian_inverse, reverse=True, **kwargs)
    _check_whole_jacobian(f, x, jacobian, **kwargs)
    _check_log_diag_jacobian(f, x, jacobian, **kwargs)


def check_gradients_not_nan(f, x, **kwargs):
    y = f(x, **kwargs)
    loss = y.mean()
    loss.backward()

    assert not any([torch.isnan(x.grad).any().item() for x in f.parameters()])


def check_area_under_pdf_1D(model, input_time=False):
    a, N = 10, 1000
    x = torch.linspace(-a, a, N).unsqueeze(-1)
    t = torch.ones_like(x) if input_time else None

    prob = model.log_prob(x, t=t).exp()
    integral = prob.sum() * 2 * a / N
    assert integral > 0.98 and integral < 1., f'Model doesn\'t define a proper 1D density (integral={integral:.5f})'

def check_area_under_pdf_2D(f, input_time=False):
    a, N = 10, 200
    x = torch.stack(torch.meshgrid(torch.linspace(-a, a, N), torch.linspace(-a, a, N)), -1).view(-1, 2)
    t = torch.ones(x.shape[0], 1) if input_time else None

    prob = f.log_prob(x, t=t).exp()
    integral = prob.sum() * (2 * a / N)**2
    assert integral > 0.98 and integral < 1., f'Model doesn\'t define a proper 2D density (integral={integral:.5f})'



import torch
import pytest
import numpy as np
import stribor as st
from stribor.test.base import *

@pytest.mark.parametrize('name', ['ELU', 'LeakyReLU'])
def test_activations(name):
    x = torch.randn(3, 4, 5)
    f = getattr(st, name)()

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)



import torch

__all__ = ['FuncAndDiagJac']


class FuncAndDiagJac(torch.autograd.Function):
    """
    Given a f: R^d -> R^d, computes both f(x) and the diagonals of the Jacobian of f(x).
    "Neural Networks with Cheap Differential Operators" (https://arxiv.org/abs/1912.03579)
    https://proceedings.neurips.cc/paper/2019/hash/770f8e448d07586afbf77bb59f698587-Abstract.html

    Args:
        exclusive_net (nn.Module): Neural network with hollow Jacobian, i.e. ith input
            does not influence the ith output. For example: `st.net.MADE`
        dimwise_net (nn.Module): Neural network with inputs `(t, x, h)` where x has `(B,1)` shape
            and `h` has `(B,H)`. It applies the f: R^{H+2} -> R function conditioned on `t` and `h`.
        t (tensor): Input time of shape (1,)
        x (tensor): Input of shape (...,D)
        flat_params (): Parameters of the exclusive and dimwise net. Can be obtained with
            `st.util.flatten_params(exclusive_net, dimwise_net)`.
        order (int): The derivation order. Default: 1
    """

    @staticmethod
    def forward(ctx, exclusive_net, dimwise_net, t, x, latent, flat_params, order=1):
        ctx.exclusive_net = exclusive_net
        ctx.dimwise_net = dimwise_net
        shape = x.shape

        with torch.enable_grad():
            t = t.detach().requires_grad_(True)
            x = x.detach().requires_grad_(True)
            if latent is not None:
                latent = latent.detach().requires_grad_(True)

            h = exclusive_net(t, x)
            x_ = x.view(-1, 1)
            h = h.contiguous().view(x_.shape[0], -1)
            h_detached = h.clone().detach().requires_grad_(True)

            if latent is not None:
                latent_ = latent.clone().unsqueeze(-2).repeat_interleave(x.shape[-1], dim=-2)
                latent_ = torch.cat([h_detached, latent_.contiguous().view(x_.shape[0], -1)], -1)
            else:
                latent_ = h_detached.clone()

            output = dimwise_net(t, x_, latent=latent_).view(*shape)

            djac = torch.autograd.grad(output.sum(), x, create_graph=True)[0]
            while order > 1:
                djac = torch.autograd.grad(djac.sum(), x, create_graph=True)[0]
                order -= 1

            ctx.save_for_backward(t, x, h, h_detached, output, djac, latent)
            return safe_detach(output), safe_detach(djac)

    @staticmethod
    def backward(ctx, grad_output, grad_djac):
        t, x, h, h_detached, output, djac, latent = ctx.saved_tensors
        grad_t = grad_x = grad_latent = grad_params = None

        f_params = list(ctx.exclusive_net.parameters()) + list(ctx.dimwise_net.parameters())

        if latent is not None:
            grad_t, grad_x, grad_h, grad_latent, *grad_params = torch.autograd.grad(
                [output, djac],
                [t, x, h_detached, latent] + f_params,
                [grad_output, grad_djac],
                retain_graph=True,
                allow_unused=True,
            )
        else:
            grad_t, grad_x, grad_h, *grad_params = torch.autograd.grad(
                [output, djac],
                [t, x, h_detached] + f_params,
                [grad_output, grad_djac],
                retain_graph=True,
                allow_unused=True,
            )

        grad_flat_params = flatten_convert_none_to_zeros(grad_params, f_params)

        if grad_h is not None:
            grad_x_from_h, *grad_params_from_h = torch.autograd.grad(
                h, [x] + f_params, grad_h, retain_graph=True, allow_unused=True
            )
            grad_x = grad_x + grad_x_from_h
            grad_flat_params = grad_flat_params + flatten_convert_none_to_zeros(grad_params_from_h, f_params)

        return None, None, grad_t, grad_x, grad_latent, grad_flat_params

def safe_detach(tensor):
    return tensor.detach().requires_grad_(tensor.requires_grad)

def flatten_convert_none_to_zeros(sequence, like_sequence):
    flat = [
        p.contiguous().view(-1) if p is not None else torch.zeros_like(q).view(-1)
        for p, q in zip(sequence, like_sequence)
    ]
    return torch.cat(flat) if len(flat) > 0 else torch.tensor([])



from typing import List
from torchtyping import TensorType

import stribor as st
import torch.nn as nn

class ResNetBlock(nn.Module):
    """
    Single ResNet block `y = x + g(x)`.

    Args:
        dim (int): Input and output size
        hidden_dims (List[int]): Hidden dimensions
        activation (str, optional): Activation function from `torch.nn`.
            Default: 'ReLU'
        final_activation (str, optional): Last activation. Default: None
    """
    def __init__(self, dim, hidden_dims, activation='ReLU', final_activation=None, **kwargs):
        super().__init__()
        self.net = st.net.MLP(dim, hidden_dims, dim, activation, final_activation)

    def forward(self, x):
        return x + self.net(x)

class ResNet(nn.Module):
    """
    ResNet - neural network consisting of residual layers.
    "Deep Residual Learning for Image Recognition" (https://arxiv.org/abs/1512.03385)

    Args:
        dim (int): Input and output size
        hidden_dims (List[int]): Hidden dimensions
        num_layers (int): Number of layers
        activation (str, optional): Activation function from `torch.nn`.
            Default: 'ReLU'
        final_activation (str, optional): Last activation. Default: None
    """
    def __init__(
        self,
        dim: int,
        hidden_dims: List[int],
        num_layers: int,
        activation: str = 'ReLU',
        final_activation: str = None,
        **kwargs,
    ):
        super().__init__()
        layers = []
        for _ in range(num_layers):
            layers.append(ResNetBlock(dim, hidden_dims, activation, final_activation))
        self.net = nn.Sequential(*layers)

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return self.net(x)



import torch
import pytest
import numpy as np
import stribor as st
from stribor.test.base import *

@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('dims', [[-1]])
def test_flip(input_shape, dims):
    torch.manual_seed(123)

    f = st.Flip(dims)
    x = torch.randn(*input_shape)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)

@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
def test_permute(input_shape):
    torch.manual_seed(123)

    f = st.Permute(input_shape[-1])
    x = torch.randn(*input_shape)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)



import pytest
import torch
import stribor as st
from stribor.test.base import *

@pytest.mark.parametrize('input_dim', [(10, 2), (2, 10), (1, 10, 2), (5, 10, 2)])
@pytest.mark.parametrize('hidden_dims', [[], [32, 64]])
@pytest.mark.parametrize('num_layers', [1, 3])
def test_resnet(input_dim, hidden_dims, num_layers):
    torch.manual_seed(123)

    dim = input_dim[-1]
    model = st.net.ResNet(dim, hidden_dims, num_layers)

    x = torch.randn(*input_dim)
    y = model(x)
    assert x.shape == y.shape


@pytest.mark.parametrize('input_dim', [(10, 2), (2, 10), (1, 10, 2), (5, 10, 2)])
@pytest.mark.parametrize('hidden_dims', [[], [32, 64]])
@pytest.mark.parametrize('num_layers', [1])
def test_invertible_resnet(input_dim, hidden_dims, num_layers):
    torch.manual_seed(123)

    dim = input_dim[-1]
    f = st.IResNet(dim, hidden_dims)

    for _ in range(5): # Run a few times so spectral_norm kicks in
        x = torch.randn(*input_dim)
        y = f(x)

    x = torch.randn(*input_dim)
    check_inverse_transform(f, x)


@pytest.mark.parametrize('input_dim', [(10, 2), (2, 10), (1, 10, 2), (5, 10, 2)])
@pytest.mark.parametrize('hidden_dims', [[], [32, 64]])
@pytest.mark.parametrize('num_layers', [1])
@pytest.mark.parametrize('time_net', ['TimeTanh', 'TimeFourierBounded'])
def test_continuous_invertible_resnet(input_dim, hidden_dims, num_layers, time_net):
    torch.manual_seed(123)

    dim = input_dim[-1]
    f = st.ContinuousIResNet(
        dim,
        hidden_dims,
        time_net=getattr(st.net, time_net)(dim, hidden_dim=8),
        n_power_iterations=10
    )

    x = torch.randn(*input_dim)
    t = torch.rand(*input_dim[:-1], 1) * 10
    for _ in range(10):
        y = f(x, t=t)

    x = torch.randn(*input_dim)
    check_inverse_transform(f, x, t=t)

    t = torch.zeros(*input_dim[:-1], 1)
    y = f(x, t=t)
    assert torch.allclose(x, y)



from typing import Optional
from torchtyping import TensorType

import math
import torch
import torch.nn.functional as F
from torch.distributions import constraints

from stribor import ElementwiseTransform


class ELU(ElementwiseTransform):
    """
    Exponential linear unit.
    Adapted from Pyro (https://pyro.ai).
    """

    bijective = True
    domain: constraints.real
    codomain: constraints.positive

    def forward(
        self,
        x: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim']:
        return F.elu(x)

    def inverse(
        self,
        y: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim']:
        zero = torch.zeros_like(y)
        log_term = torch.log1p(y)
        x = torch.max(y, zero) + torch.min(log_term, zero)
        return x

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']] = None,
    ) -> TensorType[..., 1]:
        log_diag_jacobian = self.log_diag_jacobian(x, y)
        return log_diag_jacobian.sum(-1, keepdim=True)

    def jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim', 'dim']:
        return torch.diag_embed(self.log_diag_jacobian(x, y).exp())

    def log_diag_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim']:
        return -F.relu(-x)


class LeakyReLU(ElementwiseTransform):
    """
    Leaky ReLU.
    For `x >= 0` returns `x`, else returns `negative_slope * x`.
    Adapted from Pyro (https://pyro.ai).

    Args:
        negative_slope (float): Controls the angle of the negative slope. Default: 0.01
    """

    bijective = True
    domain: constraints.real
    codomain: constraints.real

    def __init__(self, negative_slope: float = 0.01, **kwargs):
        super().__init__()
        assert negative_slope > 0, '`negative_slope` must be positive'
        self.negative_slope = negative_slope

    def _leaky_relu(
        self,
        x: TensorType[..., 'dim'],
        negative_slope: float,
    ) -> TensorType[..., 'dim']:
        zeros = torch.zeros_like(x)
        y = torch.max(zeros, x) + negative_slope * torch.min(zeros, x)
        return y

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return self._leaky_relu(x, self.negative_slope)

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return self._leaky_relu(y, 1 / self.negative_slope)

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']] = None,
    ) -> TensorType[..., 1]:
        return self.log_diag_jacobian(x, y).sum(-1, keepdim=True)

    def jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim', 'dim']:
        return torch.diag_embed(self.log_diag_jacobian(x, y).exp())

    def log_diag_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim']:
        left = torch.zeros_like(x)
        right = torch.ones_like(x) * math.log(self.negative_slope)
        return torch.where(x >= 0., left, right)



import stribor as st
import torch

def test_permuation_equivariance():
    torch.manual_seed(123)

    net = st.net.EquivariantLayer(2, 4)
    x = torch.randn(3, 5, 2)

    y1 = net(x)
    y2 = net(torch.flip(x, [1]))

    assert torch.isclose(y1, torch.flip(y2, [1])).all()



import torch
import pytest
import numpy as np
import stribor as st
from stribor.test.base import *


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('scalar_param', [True, False])
def test_fixed_affine(input_shape, scalar_param):
    torch.manual_seed(123)
    np.random.seed(123)

    dim = input_shape[-1]

    x = torch.randn(*input_shape)

    scale = np.random.rand() if scalar_param else torch.rand(dim)
    shift = np.random.normal() if scalar_param else torch.randn(dim)

    f = st.Affine(dim, scale=scale, shift=shift)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('latent_dim', [1, 13])
def test_latent_affine(input_shape, latent_dim):
    torch.manual_seed(123)
    dim = input_shape[-1]

    x = torch.randn(*input_shape)
    latent = torch.randn(*input_shape[:-1], latent_dim)

    f = st.Affine(dim, latent_net=st.net.MLP(latent_dim, [32], 2 * dim))

    check_inverse_transform(f, x, latent=latent)
    check_log_jacobian_determinant(f, x, latent=latent)
    check_gradients_not_nan(f, x, latent=latent)



@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
def test_lu_affine(input_shape):
    torch.manual_seed(123)
    dim = input_shape[-1]

    x = torch.randn(*input_shape)

    f = st.AffineLU(dim)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)
    check_gradients_not_nan(f, x)


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
def test_matrix_exponential(input_shape):
    torch.manual_seed(123)
    dim = input_shape[-1]

    x = torch.randn(*input_shape)

    f = st.MatrixExponential(dim)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)
    check_gradients_not_nan(f, x)



import pytest
import stribor as st
import torch
import torch.nn as nn


@pytest.mark.parametrize('input_dim', [(10, 2), (2, 10), (5, 10, 2), (23, 17, 7), (3, 7, 11, 13)])
def test_exact_divergence_from_jacobian(input_dim):
    torch.manual_seed(123)

    x = torch.rand(*input_dim).requires_grad_(True)

    # Linear function
    f1 = lambda x: x
    div1 = st.util.divergence_from_jacobian(f1, x)
    assert torch.isclose(div1, torch.ones_like(x)).all(), 'Incorrect divergence from Jacobian of linear function'
    div1 = st.util.divergence_exact(f1(x), x)
    assert torch.isclose(div1, torch.ones_like(x)).all(), 'Incorrect divergence from Jacobian of linear function'

    # Polynomial
    f2 = lambda x: x**3 - 4 * x**2 + 7 * x - 14
    div2 = st.util.divergence_from_jacobian(f2, x)
    assert torch.isclose(div2, 3 * x**2 - 8 * x + 7).all(), 'Incorrect divergence from Jacobian of polynomial function'
    div2 = st.util.divergence_exact(f2(x), x)
    assert torch.isclose(div2, 3 * x**2 - 8 * x + 7).all(), 'Incorrect exact divergence of polynomial function'

    # Linear layer
    f3 = nn.Linear(input_dim[-1], input_dim[-1])
    div3 = st.util.divergence_from_jacobian(f3, x)
    assert torch.isclose(div3, torch.diagonal(f3.weight).expand(input_dim)).all(), 'Incorrect divergence from Jacobian of linear layer'
    div3 = st.util.divergence_exact(f3(x), x)
    assert torch.isclose(div3, torch.diagonal(f3.weight).expand(input_dim)).all(), 'Incorrect exact divergence of linear layer'

    # Deep set
    f4 = lambda x: x**2 + (x**3).sum(-2, keepdim=True).repeat_interleave(x.shape[-2], dim=-2)
    div4 = st.util.divergence_from_jacobian(f4, x)
    assert torch.isclose(div4, 2 * x + 3 * x**2).all(), 'Incorrect divergence from Jacobian of deepset operation'
    div4 = st.util.divergence_exact_for_sets(f4(x), x)
    assert torch.isclose(div4, 2 * x + 3 * x**2).all(), 'Incorrect exact divergence of deepset operation'



from typing import Union, Optional, Tuple
from torchtyping import TensorType
from numbers import Number

import torch
import torch.nn as nn

from stribor import Transform, ElementwiseTransform


class Affine(ElementwiseTransform):
    """
    Affine flow `y = a * x + b` where `a` and `b` are vectors and operations
    are applied elementwise.

    Example:
    >>> torch.manual_seed(123)
    >>> dim, latent_dim = 2, 50
    >>> f = st.Affine(dim, st.net.MLP(latent_dim, [64, 64], 2 * dim))
    >>> f(torch.ones(1, dim), latent=torch.ones(1, latent_dim))
    (tensor([[0.7575, 0.9410]], tensor([[-0.1745, -0.1350]])

    Args:
        dim (int): Dimension of data
        latent_net (nn.Module): Neural network that maps `[..., latent]` to `[..., 2 * dim]`
        scale (tensor): Scaling coefficient `a`
        shift (tensor): Shift coefficient `b`
    """
    def __init__(
        self,
        dim: int,
        *,
        latent_net: Optional[nn.Module] = None,
        scale: Union[Number, TensorType['dim']] = None,
        shift: Union[Number, TensorType['dim']] = None,
        **kwargs,
    ):
        super().__init__()

        self.latent_net = latent_net

        if latent_net is None:
            if scale is None:
                self.log_scale = nn.Parameter(torch.empty(1, dim))
                self.shift = nn.Parameter(torch.empty(1, dim))
                nn.init.xavier_uniform_(self.log_scale)
                nn.init.xavier_uniform_(self.shift)
            else:
                if isinstance(scale, Number):
                    scale = torch.Tensor([scale])
                    shift = torch.Tensor([shift])

                assert torch.all(scale > 0), '`scale` mush have positive values'
                self.log_scale = scale.log()
                self.shift = shift

    def _get_params(
        self,
        latent: Optional[TensorType[..., 'latent']],
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 'dim']]:
        if self.latent_net is None:
            return self.log_scale, self.shift
        else:
            log_scale, shift = self.latent_net(latent).chunk(2, dim=-1)
            return log_scale, shift

    def forward(
        self,
        x: TensorType[..., 'dim'],
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        y, _ = self.forward_and_log_det_jacobian(x, latent)
        return y

    def inverse(
        self,
        y: TensorType[..., 'dim'],
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        x, _ = self.inverse_and_log_det_jacobian(y, latent)
        return x

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']] = None,
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 1]:
        _, log_det_jacobian = self.forward_and_log_det_jacobian(x, latent)
        return log_det_jacobian

    def forward_and_log_det_jacobian(self,
        x: TensorType[..., 'dim'],
        latent: Optional[TensorType[..., 'latent']] = None,
        *,
        reverse: bool = False,
        **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        log_scale, shift = self._get_params(latent)
        if reverse:
            y = (x - shift) * torch.exp(-log_scale)
        else:
            y = x * torch.exp(log_scale) + shift
        return y, log_scale.expand_as(x).sum(-1, keepdim=True)

    def inverse_and_log_det_jacobian(self, y, latent=None, **kwargs):
        x, log_det_jacobian = self.forward_and_log_det_jacobian(y, latent, reverse=True)
        return x, -log_det_jacobian

    def log_diag_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']] = None,
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        log_scale, _ = self._get_params(latent)
        return log_scale.expand_as(x)


class AffineLU(Transform):
    """
    Invertible linear layer `Wx+b` where `W=LU` is LU factorized.

    Args:
        dim: Dimension of input data
    """
    def __init__(self, dim: int, **kwargs):
        super().__init__()

        self.diag_ones = torch.eye(dim)

        self.weight = nn.Parameter(torch.empty(dim, dim))
        self.log_diag = nn.Parameter(torch.empty(1, dim))
        self.bias = nn.Parameter(torch.empty(1, dim))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.weight)
        nn.init.xavier_uniform_(self.log_diag)
        nn.init.xavier_uniform_(self.bias)

    @property
    def L(self) -> TensorType['dim', 'dim']:
        return torch.tril(self.weight, -1) + self.diag_ones

    @property
    def U(self) -> TensorType['dim', 'dim']:
        return torch.triu(self.weight, 1) + self.diag_ones * self.log_diag.exp()

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return x @ (self.L @ self.U) + self.bias

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        x = y - self.bias
        x = torch.linalg.solve_triangular(self.U, x, upper=True, left=False)
        x = torch.linalg.solve_triangular(self.L, x, upper=False, left=False)
        return x

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']] = None,
        **kwargs,
    ) -> TensorType[..., 1]:
        return self.log_diag.expand_as(x).sum(-1, keepdim=True)

    def jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'dim', 'dim']:
        return ((self.L @ self.U).T).expand(*x.shape[:-1], -1, -1)


class MatrixExponential(Transform):
    """
    Matrix exponential transformation `y = exp(W * t) @ x`.
    Corresponds to a solution of the linear ODE `dx/dt = W @ x`.

    Example:
    >>> torch.manual_seed(123)
    >>> f = stribor.MatrixExponential(2)
    >>> x = torch.rand(1, 2)
    >>> f(x, t=1.)
    (tensor([[0.0798, 1.3169]], tensor([[-0.4994,  0.4619]])
    >>> f(x, t=torch.ones(1, 1))
    (tensor([[0.0798, 1.3169]], tensor([[-0.4994,  0.4619]])

    Args:
        dim (int): Dimension of data
    """
    def __init__(self, dim, **kwargs):
        super().__init__()

        self.weight = nn.Parameter(torch.empty(dim, dim))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.weight)

    def get_time(self, t, shape):
        if isinstance(t, Number):
            t = torch.ones(*shape[:-1], 1) * t
        t = t.unsqueeze(-1)
        return t

    def forward(
        self,
        x: TensorType[..., 'dim'],
        t: Union[Number, TensorType[..., 1]] = 1.0,
        *,
        reverse: bool = False,
        **kwargs
    ) -> TensorType[..., 'dim']:
        if reverse:
            t = -t

        t = self.get_time(t, x.shape)

        W = torch.matrix_exp(self.weight * t)
        y = W @ x.unsqueeze(-1)
        y = y.squeeze(-1)

        return y

    def inverse(
        self,
        y: TensorType[..., 'dim'],
        t: Union[Number, TensorType[..., 1]] = 1.0,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        return self.forward(y, t=t, reverse=True)

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']] = None,
        t: TensorType[..., 'dim'] = 1.0,
        **kwargs,
    ) -> TensorType[..., 1]:
        t = self.get_time(t, x.shape)
        ldj = (self.weight * t).diagonal(dim1=-2, dim2=-1)
        return ldj.expand_as(x).sum(-1, keepdim=True)

    def jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        t: TensorType[..., 'dim'] = 1.0,
        **kwargs,
    ) -> TensorType[..., 'dim', 'dim']:
        t = self.get_time(t, x.shape)
        W = torch.matrix_exp(self.weight * t)
        return W.expand(*x.shape[:-1], -1, -1)



import numpy as np
import torch
import torch.nn as nn

def divergence_exact(output, input):
    diag = torch.zeros_like(input)
    for i in range(input.shape[-1]):
        diag[...,i] += torch.autograd.grad(output[...,i].sum(), input, create_graph=True)[0].contiguous()[...,i].contiguous()
    return diag

def divergence_exact_for_sets(output, input):
    diag = torch.zeros_like(input)
    for i in range(input.shape[-2]):
        for j in range(input.shape[-1]):
            diag[...,i,j] += torch.autograd.grad(output[...,i,j].sum(), input, create_graph=True)[0].contiguous()[...,i,j].contiguous()
    return diag

def divergence_approx(output, input, e, samples=1):
    out = 0
    for _ in range(samples):
        out += torch.autograd.grad(output, input, e, create_graph=True)[0] * e / samples
    return out

def divergence_from_jacobian(f, inputs):
    """
    Calculates exact divergence for any input shape.
    Best used for input-output pairs with the same shape.

    Args:
        f (callable): Function that transforms a single input or a tuple
            of inputs to an output of same size
        inputs (tensor, Tuple[tensor]): Inputs to the function
    """
    if not isinstance(inputs, tuple):
        inputs = (inputs,)

    output = f(*inputs)
    jac = torch.autograd.functional.jacobian(f, inputs, vectorize=True)

    def get_diagonal(jac, x):
        smaller_shape = output if np.prod(output.shape) < np.prod(x.shape) else x
        return torch.diag(jac.view(np.prod(x.shape), -1)).view(*smaller_shape.shape)

    div = tuple(get_diagonal(d, x) for d, x in zip(jac, inputs))
    if len(div) == 1:
        div = div[0]
    return div



from typing import List, Union, Optional, Tuple
from torchtyping import TensorType

import torch.nn as nn
import stribor as st
from stribor.net import DiffeqNet

#######################################################################################
# Nets with closed form calculated trace. Use `divergence=exact` in ContinuousTransform
#######################################################################################

class DiffeqExactTrace(nn.Module):
    """
    Differential equation with exact trace.
    Recommended to directly use tested implementations, e.g., `DiffeqExactTraceMLP`.

    Args:
        exclusive_net (st.net.DiffeqNet): Hollow Jacobian net, e.g., `st.net.DiffeqZeroTraceMLP`
        dimwise_net (st.net.DiffeqNet): Per dimension net, e.g,. `st.net.DiffeqMLP`
    """
    def __init__(
        self,
        exclusive_net: DiffeqNet,
        dimwise_net: DiffeqNet,
        return_log_det_jac: bool = True,
        **kwargs,
    ):
        super().__init__()
        self.exclusive_net = exclusive_net
        self.dimwise_net = dimwise_net
        self.return_log_det_jac = return_log_det_jac

    def forward(
        self,
        t: TensorType[1],
        x: TensorType[..., 'N', 'D'],
        latent: Optional[TensorType[..., 'N', 'L']] = None,
        **kwargs,
    ) -> Union[
        TensorType[..., 'N', 'D'],
        Tuple[TensorType[..., 'N', 'D'], TensorType[..., 'N', 'D']],
    ]:
        params = st.util.flatten_params(self.exclusive_net, self.dimwise_net)
        y, jac = st.net.FuncAndDiagJac.apply(self.exclusive_net, self.dimwise_net, t, x, latent, params)
        return (y, jac) if self.return_log_det_jac else y


class DiffeqExactTraceMLP(DiffeqExactTrace):
    """
    Exact trace network that mimics MLP architecture.

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        d_h (int): Size of conditioning vector per each dimension
        latent_dim (int, optional): Size of latent vector. Default: 0
        return_log_det_jac (bool, optional): Whether to return log-Jacobian diagonal. Default: True
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        d_h: int,
        latent_dim: int = 0,
        return_log_det_jac: bool = True,
        **kwargs,
    ):
        exclusive_net = st.net.DiffeqZeroTraceMLP(in_dim, hidden_dims, d_h * out_dim, return_log_det_jac=False, return_per_dim=True)
        dimwise_net = st.net.DiffeqMLP(d_h + latent_dim + 2, hidden_dims, 1)
        super().__init__(exclusive_net, dimwise_net, return_log_det_jac)


class DiffeqExactTraceDeepSet(DiffeqExactTrace):
    """
    Exact trace network that mimics deepset architecture.

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        d_h (int): Size of conditioning vector per each dimension
        latent_dim (int, optional): Size of latent vector. Default: 0
        pooling (str): Pooling operation. Default: 'max'
        return_log_det_jac (bool, optional): Whether to return log-Jacobian diagonal. Default: True
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        d_h: int,
        latent_dim: int = 0,
        pooling: str = 'max',
        return_log_det_jac: bool = True,
        **kwargs,
    ):
        exclusive_net = st.net.DiffeqZeroTraceDeepSet(in_dim, hidden_dims, d_h * out_dim, return_log_det_jac=False)
        dimwise_net = st.net.DiffeqMLP(d_h + latent_dim + 2, hidden_dims, 1)
        super().__init__(exclusive_net, dimwise_net, return_log_det_jac)


class DiffeqExactTraceAttention(DiffeqExactTrace):
    """
    Exact trace network that mimics deepset architecture.

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        d_h (int): Size of conditioning vector per each dimension
        latent_dim (int, optional): Size of latent vector. Default: 0
        n_heads (int, optional): Number of attention heads. Default: 1
        return_log_det_jac (bool, optional): Whether to return log-Jacobian diagonal. Default: True
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        d_h: int,
        latent_dim: int = 0,
        n_heads: int = 1,
        return_log_det_jac: bool = True,
        **kwargs,
    ):
        exclusive_net = st.net.DiffeqZeroTraceAttention(in_dim, hidden_dims, d_h * out_dim, n_heads, return_log_det_jac=False)
        dimwise_net = st.net.DiffeqMLP(d_h + latent_dim + 2, hidden_dims, 1)
        super().__init__(exclusive_net, dimwise_net, return_log_det_jac)



from typing import Optional, Tuple
from torchtyping import TensorType

import torch
import torch.nn as nn

import stribor as st
from stribor import Transform, ElementwiseTransform

class Coupling(Transform):
    """
    Coupling transformation via elementwise transforms. If `dim = 1`, set `mask` to `'none'`.
    Splits data into 2 parts based on a mask. One part generates parameters of the transform
    that will transform the rest. Efficient computation in both directions.

    Example:
    >>> import stribor as st
    >>> torch.manual_seed(123)
    >>> dim, n_bins, latent_dim = 2, 5, 50
    >>> f = st.Coupling(st.Affine(dim, st.net.MLP(dim + latent_dim, [64], 2 * dim)), mask='ordered_left_half')
    >>> f(torch.randn(1, 2), latent=torch.randn(1, latent_dim))
    tensor([[ 0.4974, -1.6288]], tensor([[0.0000, 0.3708]]
    >>> f = st.Coupling(st.Spline(dim, n_bins, st.net.MLP(dim, [64], dim * (3 * n_bins - 1))), mask='random_half')
    >>> f(torch.rand(1, dim))
    tensor([[0.9165, 0.7125]], tensor([[0.6281, -0.0000]]

    Args:
        transform (Transform): Elementwise transform with `latent_net` property.
            Latent network takes input of size `dim` and returns the parameters of the transform.
        mask (str): Mask name from `stribor.util.mask`.
            Options: `none`, `ordered_right_half` (right transforms left), `ordered_left_half`, `random_half`,
            `parity_even` (even indices transform odd), `parity_odd`.
        set_data (bool): Whether data has shape (..., N, dim) and we want to be permutation invariant. Default: False
    """
    def __init__(
        self,
        transform: ElementwiseTransform,
        mask: str,
        set_data: bool = False,
        **kwargs,
    ):
        super().__init__()

        self.transform = transform
        self.mask_func = st.util.mask.get_mask(mask) # Initializes mask generator
        self.set_data = set_data

    def _get_mask(self, x):
        if self.set_data:
            *rest, N, D = x.shape
            return self.mask_func(N).unsqueeze(-1).expand(*rest, N, D).to(x)
        else:
            return self.mask_func(x.shape[-1]).expand_as(x).to(x)

    def _get_conditioning(
        self,
        x: TensorType[..., 'dim'],
        mask: TensorType[..., 'dim'],
        latent: Optional[TensorType[..., 'latent']] = None,
    ) -> TensorType[..., 'out']:
        z = x * mask
        if x.shape[-1] == 1:
            z = z * 0
        if latent is not None:
            z = torch.cat([z, latent], -1)

        return z

    def forward(self, x, latent=None, reverse=False, **kwargs):
        mask = self._get_mask(x)
        z = self._get_conditioning(x, mask, latent)

        if reverse:
            y_ = self.transform.inverse(x, latent=z, **kwargs)
        else:
            y_ = self.transform(x, latent=z, **kwargs)

        y = y_ * (1 - mask) + x * mask
        return y

    def inverse(self, y, latent=None, **kwargs):
        return self.forward(y, latent, reverse=True)

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']] = None,
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 1]:
        mask = self._get_mask(x)
        z = self._get_conditioning(x, mask, latent)

        log_diag_jacobian = self.transform.log_diag_jacobian(x, y, latent=z, **kwargs)
        return (log_diag_jacobian * (1 - mask)).sum(-1, keepdim=True)


class ContinuousAffineCoupling(Transform):
    """
    Continuous affine coupling layer. If `dim = 1`, set `mask = 'none'`.
    Similar to `Coupling` but applies only an affine transformation
    which here depends on time `t` such that it's identity map at `t = 0`.

    Example:
    >>> import stribor as st
    >>> torch.manual_seed(123)
    >>> dim = 2
    >>> f = st.ContinuousAffineCoupling(st.net.MLP(dim+1, [64], 2 * dim), st.net.TimeLinear(2 * dim), 'parity_odd')
    >>> f(torch.rand(1, 2), t=torch.rand(1, 1))
    (tensor([[0.8188, 0.4037]], tensor([[-0.0000, -0.1784]])

    Args:
        latent_net (nn.Module): Inputs concatenation of `x` and `t` (and optionally
            `latent`) and outputs affine transformation parameters (size `2 * dim`)
        time_net (stribor.net.time_net): Time embedding with the same output
            size as `latent_net`
        mask (str): Mask name from `stribor.util.mask`
            Options: `none`, `ordered_right_half` (right transforms left), `ordered_left_half`, `random_half`,
            `parity_even` (even indices transform odd), `parity_odd`.
        concatenate_time (bool): Whether to add time to input.
    """
    def __init__(
        self,
        latent_net: nn.Module,
        time_net: nn.Module,
        mask: str,
        concatenate_time: Optional[bool] = True,
        **kwargs,
    ):
        super().__init__()

        self.latent_net = latent_net
        self.mask_func = st.util.mask.get_mask(mask)
        self.time_net = time_net
        self.concatenate_time = concatenate_time

    def _get_mask(self, x):
        return self.mask_func(x.shape[-1]).expand_as(x).to(x)

    def _get_conditioning(
        self,
        x: TensorType[..., 'dim'],
        t: TensorType[..., 1],
        mask: TensorType[..., 'dim'],
        latent: Optional[TensorType[..., 'latent']] = None,
    ) -> TensorType[..., 'out']:
        z = x * mask
        if x.shape[-1] == 1:
            z = z * 0
        if latent is not None:
            z = torch.cat([z, latent], -1)
        if self.concatenate_time:
            z = torch.cat([z, t], -1)
        return z

    def forward(
        self,
        x: TensorType[..., 'dim'],
        t: TensorType[..., 1],
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        y, _ = self.forward_and_log_det_jacobian(x, t, latent)
        return y

    def inverse(
        self,
        y: TensorType[..., 'dim'],
        t: TensorType[..., 1],
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        x, _ = self.inverse_and_log_det_jacobian(y, t, latent)
        return x

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: Optional[TensorType[..., 'dim']],
        *,
        t: TensorType[..., 1],
        latent: Optional[TensorType[..., 'latent']] = None,
        **kwargs,
    ) -> TensorType[..., 1]:
        _, log_det_jacobian = self.forward_and_log_det_jacobian(x, t, latent)
        return log_det_jacobian

    def forward_and_log_det_jacobian(self,
        x: TensorType[..., 'dim'],
        t: TensorType[..., 1],
        latent: Optional[TensorType[..., 'latent']] = None,
        *,
        reverse: bool = False,
        **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        mask = self._get_mask(x)
        z = self._get_conditioning(x, t, mask, latent)

        log_scale, shift = self.latent_net(z).chunk(2, dim=-1)
        t_log_scale, t_shift = self.time_net(t).chunk(2, dim=-1)

        if reverse:
            y = (x - shift * t_shift) * torch.exp(-log_scale * t_log_scale)
        else:
            y = x * torch.exp(log_scale * t_log_scale) + shift * t_shift
        log_diag_jacobian = log_scale * t_log_scale * (1 - mask)
        y = y * (1 - mask) + x * mask

        return y, log_diag_jacobian.sum(-1, keepdim=True)

    def inverse_and_log_det_jacobian(self, y, t, latent=None, **kwargs):
        x, log_det_jacobian = self.forward_and_log_det_jacobian(y, t, latent, reverse=True)
        return x, -log_det_jacobian



from typing import List, Optional
from torchtyping import TensorType

import torch.nn as nn

class EquivariantLayer(nn.Module):
    """
    Permutation equivariant layer. Computes the transformation of each
    set element conditioned on the sum of all elements.
    Similar to "Deep Sets" (https://arxiv.org/abs/1703.06114) and
    "On Learning Sets of Symmetric Elements" (https://arxiv.org/abs/2002.08599).

    Args:
        in_dim (int): Dimension of input
        out_dim (int): Dimension of output
    """
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        **kwargs
    ):
        super().__init__()
        self.l1 = nn.Linear(in_dim, out_dim)
        self.l2 = nn.Linear(in_dim, out_dim)

    def forward(
        self,
        x: TensorType[..., 'dim'],
        mask: Optional[TensorType[..., 'dim']] = None,
        **kwargs,
    ) -> TensorType[..., 'out']:
        y1 = self.l1(x)
        y2 = self.l2(x.sum(-2, keepdim=True))

        if mask is not None:
            mask = mask[...,0,None]
            y1 = y1 * mask
            y2 = y2 * mask / mask.sum(-2, keepdim=True)
        else:
            y2 = y2 / x.shape[-2]

        return y1 + y2

class EquivariantNet(nn.Module):
    """
    Neural network with permutation equivariant layers.
    Takes sets of elements of shape (..., N, dim). Permuting the elements across
    the second to last dimension results in the same permutation on the output.
    Args similar to `st.net.MLP` but uses `EquivariantLayer` instead of `nn.Linear`.

    Example:
    >>> net = stribor.net.EquivariantNet(2, [64, 64], 4)

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        activation (str, optional): Activation function from `torch.nn`. Default: 'Tanh'
        final_activation (str, optional): Last activation. Default: None
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        activation: str = 'Tanh',
        final_activation: str = None,
        **kwargs,
    ):
        super().__init__()

        self.activation = getattr(nn, activation)()
        self.final_activation = getattr(nn, final_activation)() if final_activation else nn.Identity()

        hidden_dims = [in_dim] + hidden_dims + [out_dim]
        self.layers = []
        for in_, out_ in zip(hidden_dims[:-1], hidden_dims[1:]):
            self.layers.append(EquivariantLayer(in_, out_))
        self.layers = nn.ModuleList(self.layers)

    def forward(
        self,
        x: TensorType[..., 'dim'],
        mask: Optional[TensorType[..., 'dim']],
        **kwargs,
    ) -> TensorType[..., 'out']:
        for layer in self.layers[:-1]:
            x = layer(x, mask=mask)
            x = self.activation(x)
        x = self.layers[-1](x, mask=mask)
        x = self.final_activation(x)
        return x



from typing import Callable, List
from torchtyping import TensorType

import torch.nn as nn

class MLP(nn.Module):
    """
    Simple multi-layer neural network.

    Example:
    >>> torch.manual_seed(123)
    >>> net = stribor.net.MLP(2, [64, 64], 1)
    >>> net(torch.randn(1, 2))
    tensor([[-0.0132]], grad_fn=<AddmmBackward>)

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        activation (str, optional): Activation function from `torch.nn`.
            Default: 'Tanh'
        final_activation (str, optional): Last activation. Default: None
        nn_linear_wrapper_func (callable, optional): Wrapper function for `nn.Linear`,
            e.g. st.util.spectral_norm. Default: None
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        activation: str = 'Tanh',
        final_activation: str = None,
        nn_linear_wrapper_func: Callable = None,
        **kwargs,
    ):
        super().__init__()

        if not nn_linear_wrapper_func:
            nn_linear_wrapper_func = lambda x: x

        hidden_dims = hidden_dims[:]
        hidden_dims.append(out_dim)
        layers = [nn.Linear(in_dim, hidden_dims[0])]

        for i in range(len(hidden_dims) - 1):
            layers.append(getattr(nn, activation)())
            layers.append(nn_linear_wrapper_func(nn.Linear(hidden_dims[i], hidden_dims[i+1])))
        layers[-1].bias.data.fill_(0.0)

        if final_activation is not None:
            layers.append(getattr(nn, final_activation)())

        self.net = nn.Sequential(*layers)

    def forward(
        self,
        x: TensorType[..., 'dim'],
        **kwargs,
    ) -> TensorType[..., 'out']:
        return self.net(x)



from typing import Optional, Tuple
from torchtyping import TensorType

import torch
import torch.nn as nn

import stribor as st
from stribor import ElementwiseTransform


class Spline(ElementwiseTransform):
    """
    Spline flow (https://arxiv.org/abs/1906.04032).
    Elementwise transformation of input vector using spline functions
    defined on the interval (lower,upper).

    Code adapted from https://github.com/bayesiains/nsf

    Example:
    >>> torch.manual_seed(123)
    >>> dim, n_bins, latent_dim = 2, 5, 50
    >>> param_size = dim * (3 * n_bins - 1)
    >>> f = stribor.Spline(dim, n_bins, latent_net=stribor.net.MLP(latent_dim, [32], param_size))
    >>> f(torch.rand(1, dim), latent=torch.rand(1, latent_dim))
    (tensor([[0.0063, 0.8072]], tensor([[ 0.0348, -0.3526]])

    Args:
        dim (int): Dimension of data
        n_bins (int): Number of bins/spline knots.
        latent_net (nn.Module, optional): Neural network that takes `latent`
            and outputs transformation parameters of size `dim * (2 * n_bins + 2)`
            for cubic and `dim * (3 * n_bins - 1)` for quadratic spline. Default: None
        lower (float, optional): Lower bound of the transformation domain. Default: 0
        upper (float, optional): Upper bound of the transformation domain. Default: 1
        spline_type (str, optional): Which spline function to use.
            Options: `quadratic`, `cubic`. Default: cubic
        latent_net: Dimension of the input latent vector
    """
    def __init__(
        self,
        dim: int,
        n_bins: int,
        latent_net: Optional[nn.Module] = None,
        lower: Optional[int] = 0,
        upper: Optional[int] = 1,
        spline_type: Optional[str] ='cubic',
        **kwargs,
    ):
        super().__init__()
        self.lower = lower
        self.upper = upper
        self.dim = dim
        self.n_bins = n_bins
        self.latent_net = latent_net

        if spline_type == 'quadratic':
            self.spline = st.util.unconstrained_rational_quadratic_spline
            self.derivative_dim = n_bins - 1
        elif spline_type == 'cubic':
            self.spline = st.util.unconstrained_cubic_spline
            self.derivative_dim = 2
        else:
            raise ValueError('spline_type must be either `quadratic` or `cubic`')

        if self.latent_net is None:
            self.width = nn.Parameter(torch.empty(self.dim, n_bins))
            self.height = nn.Parameter(torch.empty(self.dim, n_bins))
            self.derivative = nn.Parameter(torch.empty(self.dim, self.derivative_dim))
            self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.width)
        nn.init.xavier_uniform_(self.height)
        nn.init.xavier_uniform_(self.derivative)

    def _get_params(
        self, latent: TensorType[..., 'latent'] = None,
    ) -> Tuple[TensorType[..., 'dim', 'bins'], TensorType[..., 'dim', 'bins'], TensorType[..., 'dim', 'der']]:
        if latent is None:
            return self.width, self.height, self.derivative
        else:
            params = self.latent_net(latent)
            params = params.view(*params.shape[:-1], self.dim, self.n_bins * 2 + self.derivative_dim)
            width = params[...,:self.n_bins]
            height = params[...,self.n_bins:2*self.n_bins]
            derivative = params[...,2*self.n_bins:]
            return width, height, derivative

    def forward(
        self, x: TensorType[..., 'dim'], latent: TensorType[..., 'latent'] = None, **kwargs,
    ) -> TensorType[..., 'dim']:
        y, _ = self.forward_and_log_diag_jacobian(x, latent)
        return y

    def inverse(
        self, y: TensorType[..., 'dim'], latent: TensorType[..., 'latent'] = None, **kwargs,
    ) -> TensorType[..., 'dim']:
        x, _ = self.inverse_and_log_diag_jacobian(y, latent)
        return x

    def forward_and_log_diag_jacobian(
        self, x: TensorType[..., 'dim'], latent: TensorType[..., 'latent'] = None, *, reverse=False, **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 'dim']]:
        w, h, d = self._get_params(latent)
        return self.spline(x, w, h, d, inverse=reverse, lower=self.lower, upper=self.upper)

    def inverse_and_log_diag_jacobian(
        self, y: TensorType[..., 'dim'], latent: TensorType[..., 'latent'] = None, **kwargs
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 'dim']]:
        x, log_diag_jacobian = self.forward_and_log_diag_jacobian(y, latent, reverse=True)
        return x, log_diag_jacobian

    def forward_and_log_det_jacobian(
        self, x: TensorType[..., 'dim'], latent: TensorType[..., 'latent'] = None, **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        y, log_diag_jacobian = self.forward_and_log_diag_jacobian(x, latent)
        return y, log_diag_jacobian.sum(-1, keepdim=True)

    def inverse_and_log_det_jacobian(
        self, y: TensorType[..., 'dim'], latent: TensorType[..., 'latent'] = None, **kwargs,
    ) -> Tuple[TensorType[..., 'dim'], TensorType[..., 1]]:
        x, log_diag_jacobian = self.forward_and_log_diag_jacobian(y, latent, reverse=True)
        return x, log_diag_jacobian.sum(-1, keepdim=True)

    def log_det_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        latent: TensorType[..., 'latent'] = None,
        **kwargs,
    ) -> TensorType[..., 1]:
        _, log_det_jacobian = self.forward_and_log_det_jacobian(x, latent)
        return log_det_jacobian

    def log_diag_jacobian(
        self,
        x: TensorType[..., 'dim'],
        y: TensorType[..., 'dim'],
        latent: TensorType[..., 'latent'] = None,
        **kwargs,
    ) -> TensorType[..., 'dim']:
        _, log_diag_jacobian = self.forward_and_log_diag_jacobian(x, latent)
        return log_diag_jacobian



from typing import List
from torchtyping import TensorType

import torch
import torch.nn as nn
import torch.nn.functional as F

from stribor import ElementwiseTransform


class Flip(ElementwiseTransform):
    """
    Flips the order of indices along selected dims.

    Example:
    >>> f = stribor.Flip()
    >>> x = torch.tensor([[1, 2], [3, 4]])
    >>> f(x)[0]
    tensor([[2, 1],
            [4, 3]])
    >>> f = stribor.Flip([0, 1])
    >>> f(x)[0]
    tensor([[4, 3],
            [2, 1]])

    Args:
        dims (List[int]): Dimensions along which to flip the order of values.
            Default: [-1]
    """
    def __init__(self, dims: List[int] = [-1]):
        super().__init__()
        self.dims = dims

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return torch.flip(x, self.dims)

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return torch.flip(y, self.dims)

    def log_det_jacobian(self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 1]:
        return torch.zeros_like(x[...,:1])

    def log_diag_jacobian(self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return torch.eye(x.shape[-1]).flip(self.dims).diag().log().expand_as(x)


class Permute(ElementwiseTransform):
    """
    Permutes indices along the last dimension.

    Example:
    >>> torch.manual_seed(123)
    >>> f = stribor.Permute(3)
    >>> f(torch.tensor([1, 2, 3]))
    (tensor([2, 3, 1]), tensor([0, 0, 0]))
    >>> f.inverse(torch.tensor(tensor([2, 3, 1])))
    (tensor([1, 2, 3]), tensor([0, 0, 0]))

    Args:
        dim (int): Dimension of data
    """
    def __init__(self, dim: int):
        super().__init__()
        self.dim = dim
        self.permutation = torch.randperm(dim)

        self.inverse_permutation = torch.empty(dim).long()
        self.inverse_permutation[self.permutation] = torch.arange(dim)

    def forward(self, x: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        y = x[..., self.permutation]
        return y

    def inverse(self, y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        x = y[..., self.inverse_permutation]
        return x

    def log_det_jacobian(self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 1]:
        return torch.zeros_like(x[...,:1])

    def log_diag_jacobian(self, x: TensorType[..., 'dim'], y: TensorType[..., 'dim'], **kwargs) -> TensorType[..., 'dim']:
        return torch.eye(self.dim)[self.permutation].diag().log().expand_as(x)



from typing import Union, List, Optional, Tuple
from torchtyping import TensorType

import torch
import torch.nn as nn

import stribor as st
from stribor.net import DiffeqNet

###################################################################
# Volume preserving, zero trace nets. Use `divergence=exact` in CNF
###################################################################

class DiffeqZeroTraceMLP(DiffeqNet):
    """
    Zero trace MLP transformation based on MADE network.

    Example:
    >>> dim = 3
    >>> net = stribor.net.DiffeqZeroTraceMLP(dim, [64], dim, return_log_det_jac=False)
    >>> x = torch.randn(32, dim)
    >>> t = torch.rand(32, 1)
    >>> net(t, x).shape
    torch.Size([32, 3])

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        return_log_det_jac (bool, optional): Whether to return the log-Jacobian
            diagonal values (always zero). Default: True
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        return_log_det_jac: Optional[bool] = True,
        **kwargs,
    ):
        super().__init__()
        self.return_log_det_jac = return_log_det_jac
        self.net1 = st.net.MADE(in_dim, hidden_dims, out_dim, natural_ordering=True,
                                reverse_ordering=False, return_per_dim=True)
        self.net2 = st.net.MADE(in_dim, hidden_dims, out_dim, natural_ordering=True,
                                reverse_ordering=True, return_per_dim=True)

    def forward(
        self, t: TensorType[..., 1], x: TensorType[..., 'dim'], **kwargs,
    ) -> Union[
        TensorType[..., 'out'],
        Tuple[TensorType[..., 'out'], TensorType[..., 'dim']]
    ]:
        y = self.net1(x, **kwargs) + self.net2(x, **kwargs)
        y = y.reshape(*y.shape[:-2], -1)
        return (y, torch.zeros_like(x)) if self.return_log_det_jac else y


def exclusive_sum_pooling(
    x: TensorType[..., 'N', 'D'], mask: Union[TensorType[..., 'N', 1], TensorType[..., 'N', 'D']],
) -> TensorType[..., 'N', 'D']:
    emb = x.sum(-2, keepdims=True)
    return emb - x

def exclusive_mean_pooling(
    x: TensorType[..., 'N', 'D'], mask: Union[TensorType[..., 'N', 1], TensorType[..., 'N', 'D']],
) -> TensorType[..., 'N', 'D']:
    emb = exclusive_sum_pooling(x, mask)
    N = mask.sum(-2, keepdim=True)
    y = emb / torch.max(N - 1, torch.ones_like(N))[0]
    return y

def exclusive_max_pooling(
    x: TensorType[..., 'N', 'D'], mask: Union[TensorType[..., 'N', 1], TensorType[..., 'N', 'D']],
) -> TensorType[..., 'N', 'D']:
    if x.shape[-2] == 1: # If only one element in set
        return torch.zeros_like(x)

    first, second = torch.topk(x, 2, dim=-2).values.chunk(2, dim=-2)
    indicator = (x == first).float()
    y = (1 - indicator) * first + indicator * second
    return y


class ZeroTraceEquivariantEncoder(nn.Module):
    """
    Pooling layer with zero trace Jacobian.

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        pooling (str): Which type of pooling to use. Options: 'mean', 'max', 'sum'
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        pooling: str,
        **kwargs,
    ):
        super().__init__()
        self.pooling = pooling
        self.in_dim = in_dim
        self.set_emb = st.net.DiffeqMLP(in_dim + 1, hidden_dims, out_dim)

    def forward(
        self,
        t: TensorType[..., 1],
        x: TensorType[..., 'N', 'D'],
        mask: Optional[Union[TensorType[..., 'N', 1], TensorType[..., 'N', 'D']]] = None,
        **kwargs,
    ) -> TensorType[..., 'N', 'D', 'O']:
        if mask is None:
            mask = torch.ones(*x.shape[:-1], 1)
        else:
            mask = mask[...,0,None]

        x = self.set_emb(t, x) * mask
        if self.pooling == 'mean':
            y = exclusive_mean_pooling(x, mask)
        elif self.pooling == 'max':
            y = exclusive_max_pooling(x, mask)
        elif self.pooling == 'sum':
            y = exclusive_sum_pooling(x, mask)
        y = y.unsqueeze(-2).repeat_interleave(self.in_dim, dim=-2)
        return y


class DiffeqZeroTraceDeepSet(DiffeqNet):
    """
    Zero trace deepset transformation.

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        pooling (str, optional): Which pooling to use. Default: 'max'
        return_log_det_jac (bool, optional): Whether to return the log-Jacobian
            diagonal values (always zero). Default: True
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        pooling: str = 'max',
        return_log_det_jac: Optional[bool] = True,
        **kwargs,
    ):
        super().__init__()
        self.elementwise = st.net.MADE(in_dim, hidden_dims, out_dim, return_per_dim=True)
        self.interaction = ZeroTraceEquivariantEncoder(in_dim, hidden_dims, out_dim // in_dim, pooling)
        self.return_log_det_jac = return_log_det_jac

    def forward(
        self,
        t: TensorType[..., 1],
        x: TensorType[..., 'N', 'D'],
        mask: Optional[Union[TensorType[..., 'N', 1], TensorType[..., 'N', 'D']]] = None,
        latent: Optional[TensorType[...]] = None,
        **kwargs,
    ) -> Union[
        TensorType[..., 'N', 'O'],
        Tuple[TensorType[..., 'N', 'O'], TensorType[..., 'N', 'D']]
    ]:
        div = torch.zeros_like(x)
        if latent is not None:
            x = torch.cat([x, latent], -1)
        if mask is None:
            mask = torch.ones(*x.shape[:-1], 1).to(x)
        y = self.elementwise(x) + self.interaction(t, x, mask=mask)
        y = y * mask.unsqueeze(-1)
        y = y.reshape(*y.shape[:-2], -1)
        return (y, div) if self.return_log_det_jac else y


class DiffeqZeroTraceAttention(DiffeqNet):
    """
    Zero trace attention transformation.

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions
        out_dim (int): Output size
        n_heads (int, optional): Number of heads in multihead attention. Default: 1
        return_log_det_jac (bool, optional): Whether to return the log-Jacobian
            diagonal values (always zero). Default: True
    """
    def __init__(
        self,
        in_dim: int,
        hidden_dims: List[int],
        out_dim: int,
        n_heads: int = 1,
        return_log_det_jac: Optional[bool] = True,
        **kwargs,
    ):
        super().__init__()
        self.n_heads = n_heads
        self.return_log_det_jac = return_log_det_jac

        self.q = st.net.MADE(in_dim, hidden_dims[:-1], hidden_dims[-1] * in_dim, return_per_dim=True)
        self.k = st.net.MLP(in_dim, hidden_dims[:-1], hidden_dims[-1])
        self.v = st.net.MLP(in_dim, hidden_dims[:-1], hidden_dims[-1])
        self.proj = st.net.MLP(hidden_dims[-1], [], out_dim // in_dim)

    def forward(self,
        t: TensorType[..., 1],
        x: TensorType[..., 'N', 'D'],
        mask: Optional[TensorType[..., 'N', 'D']] = None,
        latent: Optional[TensorType[...]] = None,
        **kwargs,
    ) -> Union[
        TensorType[..., 'N', 'O'],
        Tuple[TensorType[..., 'N', 'O'], TensorType[..., 'N', 'D']]
    ]:
        query = self.q(x).transpose(-2, -3) # (B, N, D) -> (B, D, N, H)
        key = self.k(x).unsqueeze(-2).repeat_interleave(x.shape[-1], dim=-2).transpose(-2, -3) # (B, D, N, H)
        value = self.v(x).unsqueeze(-2).repeat_interleave(x.shape[-1], dim=-2).transpose(-2, -3) # (B, D, N, H)

        y = st.net.attention(query, key, value, self.n_heads, True, mask) # (B, D, N, H)
        y = y.transpose(-2, -3) # (B, N, D, H)
        y = self.proj(y).view(*y.shape[:-2], -1) # (B, N, D, O) -> (B, N, D * O)

        return (y, torch.zeros_like(x)) if self.return_log_det_jac else y



import torch
import pytest
import stribor as st
from stribor.test.base import *


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('latent_dim', [0, 1, 13])
def test_affine_coupling(input_shape, latent_dim):
    torch.manual_seed(123)
    dim = input_shape[-1]

    x = torch.randn(*input_shape)
    if latent_dim == 0:
        latent = None
    else:
        latent = torch.randn(*input_shape[:-1], latent_dim)

    f = st.Coupling(
        transform=st.Affine(dim, latent_net=st.net.MLP(dim + latent_dim, [13], 2 * dim)),
        mask='ordered_left_half',
    )

    check_inverse_transform(f, x, latent=latent)
    check_log_jacobian_determinant(f, x, latent=latent)
    check_gradients_not_nan(f, x, latent=latent)


@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
@pytest.mark.parametrize('latent_dim', [0, 1, 13])
def test_continuous_affine_coupling(input_shape, latent_dim):
    torch.manual_seed(123)
    dim = input_shape[-1]

    x = torch.randn(*input_shape)
    if latent_dim == 0:
        latent = None
    else:
        latent = torch.randn(*input_shape[:-1], latent_dim)

    t = torch.randn_like(x[...,:1])

    f = st.ContinuousAffineCoupling(
        latent_net=st.net.MLP(dim + 1 + latent_dim, [13], 2 * dim),
        time_net=st.net.TimeLinear(2 * dim),
        mask='ordered_left_half',
    )

    check_inverse_transform(f, x, t=t, latent=latent)
    check_log_jacobian_determinant(f, x, t=t, latent=latent)
    check_gradients_not_nan(f, x, t=t, latent=latent)



from setuptools import setup, find_packages

install_requires = [
    'numpy',
    'pytest',
    'scipy',
    'torch>=1.8.0',
    'torchdiffeq==0.2.2',
    'torchtyping==0.1.4',
]

with open('README.md', 'r') as f:
    long_description = f.read()

setup(name='stribor',
      version='0.2.0',
      description='Library for normalizing flows and neural flows',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/mbilos/stribor',
      author='Marin Bilos',
      author_email='bilos@in.tum.de',
      packages=find_packages(),
      install_requires=install_requires,
      python_requires='>=3.7',
      zip_safe=False,
)



import stribor as st
import torch

def test_ordered_right_half_mask():
    gen = st.util.get_mask('ordered_right_half')

    assert torch.eq(gen(1), torch.Tensor([1])).all()
    assert torch.eq(gen(2), torch.Tensor([0, 1])).all()
    assert torch.eq(gen(5), torch.Tensor([0, 0, 1, 1, 1])).all()

def test_ordered_left_half_mask():
    gen = st.util.get_mask('ordered_left_half')

    assert torch.eq(gen(1), torch.Tensor([1])).all()
    assert torch.eq(gen(2), torch.Tensor([1, 0])).all()
    assert torch.eq(gen(5), torch.Tensor([1, 1, 0, 0, 0])).all()

def test_parity_even_mask():
    gen = st.util.get_mask('parity_even')

    assert torch.eq(gen(1), torch.Tensor([1])).all()
    assert torch.eq(gen(2), torch.Tensor([0, 1])).all()
    assert torch.eq(gen(5), torch.Tensor([0, 1, 0, 1, 0])).all()

def test_parity_odd_mask():
    gen = st.util.get_mask('parity_odd')

    assert torch.eq(gen(1), torch.Tensor([1])).all()
    assert torch.eq(gen(2), torch.Tensor([1, 0])).all()
    assert torch.eq(gen(5), torch.Tensor([1, 0, 1, 0, 1])).all()



import torch
import pytest
import numpy as np
import stribor as st
from stribor.test.base import *

@pytest.mark.parametrize('input_shape', [(1, 1), (2, 10), (10, 2), (7, 4, 5)])
def test_sigmoid(input_shape):
    np.random.seed(123)
    torch.manual_seed(123)

    f = st.Sigmoid()
    x = torch.randn(*input_shape)

    check_inverse_transform(f, x)
    check_log_jacobian_determinant(f, x)

    g = st.Logit()
    y = torch.rand(*input_shape) * 0.5 + 0.25

    check_inverse_transform(g, y)
    check_log_jacobian_determinant(g, y)



import stribor as st
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


def attention(query, key, value, n_heads=1, mask_diagonal=False, mask=None):
    """
    Multihead attention layer with optional masking.
    "Attention Is All You Need" (https://arxiv.org/abs/1706.03762)

    Args:
        query (tensor): Query matrix (..., N, D)
        key (tensor): Key matrix (..., N, D)
        value (tensor): Value matrix (..., N, D)
        h_heads (int, optional): Number of attention heads, must divide D. Default: 1
        mask_diagonal (bool, optional): Whether to mask the diagonal in QK^T matrix,
            i.e., if self-interaction is disallowed. Default: False
        mask (tensor, optional): Mask with shape (..., N, 1), masked elements
            will be ignored. Default: None

    Returns:
        y (tensor): Result of multihead attention with shape (..., N, D)
    """

    *query_shape, _ = query.shape
    *value_shape, D = value.shape
    query = query.view(*query_shape, n_heads, D // n_heads).transpose(-2, -3)
    value = value.view(*value_shape, n_heads, D // n_heads).transpose(-2, -3)
    key = key.view(*value_shape, n_heads, D // n_heads).transpose(-2, -3)

    att = query @ key.transpose(-1, -2) * (1 / key.shape[-1])**0.5

    if mask_diagonal:
        att.masked_fill_(torch.eye(att.shape[-1]).bool(), -np.inf)
    if mask is not None:
        att_mask = 1 - mask.transpose(-1, -2).unsqueeze(-2).repeat_interleave(att.shape[-2], dim=-2)
        att.masked_fill_(att_mask.bool(), -np.inf)

    att = st.util.safe_softmax(att, -1)

    y = att @ value

    y = y.transpose(-2, -3).reshape(*query_shape, -1)

    if mask is not None and y.shape[-2] == mask.shape[-2]:
        y = y * mask
    return y

class Attention(nn.Module):
    """
    Attention layer.

    Example:
    >>> net = stribor.net.Attention(3, [64, 64], 8, n_heads=4)
    >>> x = torch.randn(32, 10, 3)
    >>> net(x, x, x).shape
    torch.Size([32, 10, 8])

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions of embedding network, last dimmension
            is used as an embedding size for queries, keys and values
        out_dim (int): Output size
        h_heads (int, optional): Number of attention heads, must divide last element
            of `hidden_dims`. Default: 1
        mask_diagonal (bool, optional): Whether to mask the diagonal in similarity matrix,
            i.e., if self-interaction is disallowed. Default: False
    """
    def __init__(self, in_dim, hidden_dims, out_dim, n_heads=1, mask_diagonal=False, **kwargs):
        super().__init__()

        self.mask_diagonal = mask_diagonal
        self.n_heads = n_heads

        self.key = st.net.MLP(in_dim, hidden_dims[:-1], hidden_dims[-1])
        self.query = st.net.MLP(in_dim, hidden_dims[:-1], hidden_dims[-1])
        self.value = st.net.MLP(in_dim, hidden_dims[:-1], hidden_dims[-1])
        self.proj = nn.Linear(hidden_dims[-1], out_dim)

    def forward(self, query, key, value, mask=None, **kwargs):
        """
        Args:
            query (tensor): Queries with shape (..., N, in_dim)
            key (tensor): Keys with shape (..., N, in_dim)
            value (tensor): Values with shape (..., N, in_dim)
            mask (tensor, optional): Queries with shape (..., N, 1)

        Returns:
            y (tensor): Outputs with shape (..., N, out_dim)
        """
        y = attention(self.query(query), self.key(key), self.value(value), self.n_heads, self.mask_diagonal, mask)
        return self.proj(y)


class SelfAttention(Attention):
    """
    Self attention layer.
    Same as `st.net.Attention` but `forward` takes only `x` and `mask`.

    Args:
        Same as in `st.net.Attention`
    """
    def __init__(self, in_dim, hidden_dim, out_dim, n_heads=1, mask_diagonal=False, **kwargs):
        super().__init__(in_dim, hidden_dim, out_dim, n_heads, mask_diagonal)

    def forward(self, x, mask=None, **kwargs):
        """
        Args:
            x (tensor): Input, acts as query, key and value, with shape (..., N, in_dim)
            mask (tensor, optional): Queries with shape (..., N, 1)

        Returns:
            y (tensor): Outputs with shape (..., N, out_dim)
        """
        return super().forward(x, x, x, mask=mask)


class InducedSelfAttention(nn.Module):
    """
    Induced self attention that used inducing points to reduce the complexity
    of computing the attention for big sets.
    "Set Transformer: A Framework for Attention-based Permutation-Invariant Neural Networks"
    (https://arxiv.org/abs/1810.00825)

    Args:
        in_dim (int): Input size
        hidden_dims (List[int]): Hidden dimensions of embedding network, last dimmension
            is used as an embedding size for queries, keys and values
        out_dim (int): Output size
        h_heads (int, optional): Number of attention heads, must divide last element
            of `hidden_dims`. Default: 1
        n_points (int, optional): Number of inducing point. Default: 32
    """
    def __init__(self, in_dim, hidden_dim, out_dim, n_heads=1, n_points=32, **kwargs):
        super().__init__()

        self.att1 = Attention(in_dim, hidden_dim, in_dim, n_heads)
        self.att2 = Attention(in_dim, hidden_dim, out_dim, n_heads)
        self.points = nn.Parameter(torch.empty(n_points, in_dim).uniform_(-1., 1.))

    def forward(self, x, mask=None, **kwargs):
        h = self.points[(None,) * (len(x.shape) - 2)].repeat(*x.shape[:-2], 1, 1)
        h = self.att1(h, x, x, mask=mask, **kwargs)
        y = self.att2(x * (1 if mask is None else mask), h, h, **kwargs)
        return y



import stribor as st
import torch

def test_normal():
    x = torch.randn(4, 3, 5, 2)
    p0 = torch.distributions.Normal(torch.zeros(2), torch.ones(2)).log_prob(x).sum(-1)
    p1 = st.Normal(torch.zeros(2), torch.ones(2)).log_prob(x)
    p2 = st.MultivariateNormal(torch.zeros(2), torch.eye(2)).log_prob(x)
    assert (p0 == p1).all() and torch.isclose(p1, p2).all()



from torchtyping import TensorType

import torch
import torch.nn as nn
import torch.nn.functional as F

class TimeIdentity(nn.Module):
    def __init__(self, out_dim: int, **kwargs):
        super().__init__()
        self.out_dim = out_dim

    def forward(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return t.repeat_interleave(self.out_dim, dim=-1)

    def derivative(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return torch.ones_like(t).repeat_interleave(self.out_dim, dim=-1)

class TimeLinear(nn.Module):
    def __init__(self, out_dim: int, **kwargs):
        super().__init__()
        self.scale = nn.Parameter(torch.randn(1, out_dim))
        nn.init.xavier_uniform_(self.scale)

    def forward(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return self.scale * t

    def derivative(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return self.scale * torch.ones_like(t)


class TimeTanh(TimeLinear):
    def forward(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return torch.tanh(self.scale * t)

    def derivative(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return self.scale * (1 - self.forward(t)**2)

class TimeLog(TimeLinear):
    def forward(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return torch.log(self.scale.exp() * t + 1)

    def derivative(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        return self.scale.exp() / (self.scale.exp() * t + 1)

class TimeFourier(nn.Module):
    """
    Fourier features: sum_k x_k sin(s_k t).

    Args:
        out_dim: Output dimension.
        hidden_dim: Number of fourier features.
        lmbd: Lambda parameter of exponential distribution used
            to initialize shift parameters. (default: 0.5)
    """
    def __init__(self, out_dim, hidden_dim, lmbd=0.5, bounded=False, **kwargs):
        super().__init__()
        self.bounded = bounded
        self.hidden_dim = hidden_dim
        self.shift = nn.Parameter(-torch.log(1 - torch.rand(out_dim, hidden_dim)) / lmbd)

        self.weight = nn.Parameter(torch.Tensor(out_dim, hidden_dim))
        nn.init.xavier_normal_(self.weight)

    def get_scale(self):
        if self.bounded:
            return F.softmax(self.weight, -1) / 2
        else:
            return self.weight / self.hidden_dim

    def forward(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        t = t.unsqueeze(-1)
        scale = self.get_scale()
        t = scale * torch.sin(self.shift * t)
        t = t.sum(-1)
        return t

    def derivative(self, t: TensorType[..., 1]) -> TensorType[..., 'out']:
        t = t.unsqueeze(-1)
        scale = self.get_scale()
        t = self.shift * scale * torch.cos(self.shift * t)
        t = t.sum(-1)
        return t

class TimeFourierBounded(TimeFourier):
    """ Same as TimeFourier but between 0 and 1 """
    def __init__(self, out_dim, hidden_dim, lmbd=0.5, **kwargs):
        super().__init__(out_dim, hidden_dim, lmbd, True)
